#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest

import octobot_trading.personal_data as personal_data
import octobot_trading.errors as errors
from tests.exchanges import backtesting_config, backtesting_exchange_manager, fake_backtesting


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def oco_group(backtesting_exchange_manager):
    orders_manager = mock.Mock()
    orders_manager.get_order_from_group = mock.Mock()
    return personal_data.OneCancelsTheOtherOrderGroup("name",  orders_manager)


def order_mock():
    order = mock.Mock()
    order.is_open = mock.Mock(return_value=True)
    order.trader = mock.Mock()
    order.trader.cancel_order = mock.AsyncMock()
    return order


async def test_on_fill(oco_group):
    order = order_mock()
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
         as get_order_from_group_mock:
        await oco_group.on_fill(order)
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_called_once_with(oco_group.name)
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order])) \
         as get_order_from_group_mock:
        await oco_group.on_fill(order, ignored_orders=["hello"])
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_called_once_with(oco_group.name)
    other_order = order_mock()
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order,
                                                                                                     other_order])) \
         as get_order_from_group_mock:
        await oco_group.on_fill(order)
        order.trader.cancel_order.assert_not_called()
        other_order.trader.cancel_order.assert_called_once_with(other_order, ignored_order=order)
        get_order_from_group_mock.assert_called_once_with(oco_group.name)


async def test_on_cancel(oco_group):
    order = order_mock()
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
         as get_order_from_group_mock:
        await oco_group.on_cancel(order)
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_called_once_with(oco_group.name)
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order])) \
         as get_order_from_group_mock:
        await oco_group.on_cancel(order, ignored_orders=["hi"])
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_called_once_with(oco_group.name)
        get_order_from_group_mock.reset_mock()
        with pytest.raises(errors.OrderGroupTriggerArgumentError):
            await oco_group.on_cancel(order, ignored_orders=["hi", "ho"])
        order.trader.cancel_order.assert_not_called()
        get_order_from_group_mock.assert_not_called()
    other_order = order_mock()
    with mock.patch.object(oco_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order,
                                                                                                     other_order])) \
         as get_order_from_group_mock:
        await oco_group.on_cancel(order, ignored_orders=["hi"])
        order.trader.cancel_order.assert_not_called()
        other_order.trader.cancel_order.assert_called_once_with(other_order, ignored_order="hi")
        get_order_from_group_mock.assert_called_once_with(oco_group.name)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import ccxt.async_support
import ccxt

import octobot_trading.enums as enums


class BasicExchangeWrapper:
    def __init__(self, exchange_class_string: str, exchange_lib: enums.ExchangeWrapperLibs):
        """
        Always call self.stop() when done with an async ccxt exchange wrapper
        :param exchange_class_string: name of the exchange to create
        :param exchange_lib: lib to use to connect to the exchange
        """
        self.exchange_class_string = exchange_class_string
        self.exchange_lib = exchange_lib
        self.exchange = None

        self._exchange_factory()

    async def get_available_time_frames(self):
        if self.exchange_lib in (enums.ExchangeWrapperLibs.CCXT, enums.ExchangeWrapperLibs.ASYNC_CCXT):
            return self.exchange.timeframes

    async def stop(self):
        if self.exchange_lib is enums.ExchangeWrapperLibs.ASYNC_CCXT:
            await self.exchange.close()

    def _exchange_factory(self):
        if self.exchange_lib is enums.ExchangeWrapperLibs.ASYNC_CCXT:
            self.exchange = getattr(ccxt.async_support, self.exchange_class_string)()
        elif self.exchange_lib is enums.ExchangeWrapperLibs.CCXT:
            self.exchange = getattr(ccxt, self.exchange_class_string)()
        else:
            raise NotImplementedError(
                f"{self.exchange_lib} exchange lib is not implemented in {self.__class__.__name__}")


@contextlib.asynccontextmanager
async def temporary_exchange_wrapper(exchange_class_string: str, exchange_lib: enums.ExchangeWrapperLibs):
    """
    Automatically call to on exchange wrapper
    :param exchange_class_string: name of the exchange to create
    :param exchange_lib: lib to use to connect to the exchange
    :return:
    """
    wrapper = BasicExchangeWrapper(exchange_class_string, exchange_lib)
    try:
        yield wrapper
    finally:
        await wrapper.stop()



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


cdef class MarginContract:
    cdef readonly str pair

    cdef public object margin_type

    cdef readonly object contract_size
    cdef readonly object current_leverage
    cdef readonly object maximum_leverage

    cdef readonly dict risk_limit

    cpdef bint is_isolated(self)
    cpdef bint check_leverage_update(self, object new_leverage)
    cpdef void set_current_leverage(self, object new_leverage)
    cpdef void set_margin_type(self, bint is_isolated=*, bint is_cross=*)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from contextlib import asynccontextmanager

import octobot_commons.constants as commons_constants
from octobot_commons.asyncio_tools import wait_asyncio_next_cycle
from octobot_commons.tests.test_config import load_test_config
import octobot_trading.api as api
import octobot_trading.exchanges as exchanges


@asynccontextmanager
async def get_exchange_manager(exchange_name, config=None):
    config = config or load_test_config()
    if exchange_name not in config[commons_constants.CONFIG_EXCHANGES]:
        config[commons_constants.CONFIG_EXCHANGES][exchange_name] = {}
    exchange_manager_instance = exchanges.ExchangeManager(config, exchange_name)
    await exchange_manager_instance.initialize()
    try:
        yield exchange_manager_instance
    finally:
        await exchange_manager_instance.stop()
        api.cancel_ccxt_throttle_task()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.util as util
cimport numpy as np
np.import_array()

cdef class CandlesManager(util.Initializable):
    cdef object logger

    cdef public bint candles_initialized
    cdef public int max_candles_count

    cdef public np.float64_t[::1] close_candles
    cdef public np.float64_t[::1] open_candles
    cdef public np.float64_t[::1] high_candles
    cdef public np.float64_t[::1] low_candles
    cdef public np.float64_t[::1] time_candles
    cdef public np.float64_t[::1] volume_candles

    cdef public int close_candles_index
    cdef public int open_candles_index
    cdef public int high_candles_index
    cdef public int low_candles_index
    cdef public int time_candles_index
    cdef public int volume_candles_index

    cdef public bint reached_max

    cpdef np.ndarray get_symbol_close_candles(self, int limit=*)
    cpdef np.ndarray get_symbol_open_candles(self, int limit=*)
    cpdef np.ndarray get_symbol_high_candles(self, int limit=*)
    cpdef np.ndarray get_symbol_low_candles(self, int limit=*)
    cpdef np.ndarray get_symbol_time_candles(self, int limit=*)
    cpdef np.ndarray get_symbol_volume_candles(self, int limit=*)

    cpdef dict get_symbol_prices(self, object limit=*)
    cpdef list get_candles(self, object limit=*)
    cpdef void add_old_and_new_candles(self, list candles_data)
    cpdef void add_new_candle(self, list new_candle_data)
    cpdef void replace_all_candles(self, list all_candles_data)

    # private
    cdef void _set_all_candles(self, object new_candles_data)
    cdef void _change_current_candle(self)
    cdef bint _should_add_new_candle(self, new_open_time)
    cdef void _check_max_candles(self)
    cdef object _inc_candle_index(self)
    cdef void _reset_candles(self)
    cdef np.ndarray _extract_limited_data(self, np.float64_t[::1] data, int limit=*, int max_limit=*)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

import pytest
from math import nan

from octobot_trading.exchange_data.ticker.ticker_manager import TickerManager
from octobot_trading.enums import ExchangeConstantsMiniTickerColumns, ExchangeConstantsTickersColumns
from tests.test_utils.random_numbers import random_price, random_quantity, random_timestamp
from tests import event_loop

pytestmark = pytest.mark.asyncio


@pytest.fixture()
async def ticker_manager():
    t_manager = TickerManager()
    await t_manager.initialize()
    return t_manager


async def test_init(ticker_manager):
    assert not ticker_manager.ticker == {}
    assert not ticker_manager.mini_ticker == {}


async def test_update_and_reset_ticker(ticker_manager):
    price = random_price()
    tm = random_timestamp()
    vlm = random_quantity()
    ticker_manager.ticker_update({
        ExchangeConstantsTickersColumns.ASK.value: price,
        ExchangeConstantsTickersColumns.ASK_VOLUME.value: vlm,
        ExchangeConstantsTickersColumns.TIMESTAMP.value: tm
    })
    assert ticker_manager.ticker[ExchangeConstantsTickersColumns.ASK.value] == price
    assert ticker_manager.ticker[ExchangeConstantsTickersColumns.ASK_VOLUME.value] == vlm
    assert ticker_manager.ticker[ExchangeConstantsTickersColumns.TIMESTAMP.value] == tm
    if not os.getenv('CYTHON_IGNORE'):
        ticker_manager.reset_ticker()
        assert ticker_manager.ticker == {
            ExchangeConstantsTickersColumns.ASK.value: nan,
            ExchangeConstantsTickersColumns.ASK_VOLUME.value: nan,
            ExchangeConstantsTickersColumns.BID.value: nan,
            ExchangeConstantsTickersColumns.BID_VOLUME.value: nan,
            ExchangeConstantsTickersColumns.OPEN.value: nan,
            ExchangeConstantsTickersColumns.LOW.value: nan,
            ExchangeConstantsTickersColumns.HIGH.value: nan,
            ExchangeConstantsTickersColumns.CLOSE.value: nan,
            ExchangeConstantsTickersColumns.LAST.value: nan,
            ExchangeConstantsTickersColumns.AVERAGE.value: nan,
            ExchangeConstantsTickersColumns.SYMBOL.value: nan,
            ExchangeConstantsTickersColumns.QUOTE_VOLUME.value: nan,
            ExchangeConstantsTickersColumns.TIMESTAMP.value: 0,
            ExchangeConstantsTickersColumns.VWAP.value: nan
        }


async def test_update_and_reset_mini_ticker(ticker_manager):
    vol = random_quantity()
    open_p = random_price()
    ticker_manager.mini_ticker_update({
            ExchangeConstantsMiniTickerColumns.HIGH_PRICE.value: random_price(),
            ExchangeConstantsMiniTickerColumns.LOW_PRICE.value: random_price(),
            ExchangeConstantsMiniTickerColumns.OPEN_PRICE.value: open_p,
            ExchangeConstantsMiniTickerColumns.CLOSE_PRICE.value: random_price(),
            ExchangeConstantsMiniTickerColumns.VOLUME.value: vol,
            ExchangeConstantsMiniTickerColumns.TIMESTAMP.value: random_timestamp()
        })
    assert ticker_manager.mini_ticker[ExchangeConstantsMiniTickerColumns.VOLUME.value] == vol
    assert ticker_manager.mini_ticker[ExchangeConstantsMiniTickerColumns.OPEN_PRICE.value] == open_p
    if not os.getenv('CYTHON_IGNORE'):
        ticker_manager.reset_mini_ticker()
        assert ticker_manager.mini_ticker == {
                ExchangeConstantsMiniTickerColumns.HIGH_PRICE.value: nan,
                ExchangeConstantsMiniTickerColumns.LOW_PRICE.value: nan,
                ExchangeConstantsMiniTickerColumns.OPEN_PRICE.value: nan,
                ExchangeConstantsMiniTickerColumns.CLOSE_PRICE.value: nan,
                ExchangeConstantsMiniTickerColumns.VOLUME.value: nan,
                ExchangeConstantsMiniTickerColumns.TIMESTAMP.value: 0
            }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.enums as channel_enums
import octobot_commons.logging as logging
import octobot_commons.enums as commons_enums
import octobot_commons.authentication as authentication
import octobot_trading.signals.channel.remote_trading_signal as signals_channel
import octobot_trading.signals.trading_signal_factory as trading_signal_factory


class RemoteTradingSignalProducer(signals_channel.RemoteTradingSignalChannelProducer):
    def __init__(self, channel, bot_id):
        super().__init__(channel)
        # the trading mode instance logger
        self.logger = logging.get_logger(self.__class__.__name__)

        # Define trading modes default consumer priority level
        self.priority_level: int = channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value

        self.bot_id = bot_id

    async def subscribe_to_product_feed(self, feed_id):
        await authentication.Authenticator.instance().register_feed_callback(commons_enums.CommunityChannelTypes.SIGNAL, self.on_new_signal,
                                                                             identifier=feed_id)

    async def on_new_signal(self, parsed_message) -> None:
        try:
            signal = trading_signal_factory.create_trading_signal(parsed_message)
            await self.send(signal, self.bot_id)
        except Exception as e:
            self.logger.exception(e, True, f"Error when processing signal: {e}")

    def flush(self) -> None:
        """
        Flush all instance objects reference
        """
        self.exchange_manager = None



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


cimport octobot_trading.util as util
cimport octobot_trading.personal_data.portfolios.portfolio_manager as portfolio_manager

cdef class HistoricalPortfolioValueManager(util.Initializable):
    cdef public object logger

    cdef public portfolio_manager.PortfolioManager portfolio_manager

    cdef public str _portfolio_type_suffix
    cdef public list saved_time_frames
    cdef public str data_source
    cdef public str version
    cdef public int max_history_size
    cdef public object historical_portfolio_value
    cdef public object run_dbs_identifier

    cpdef object get_historical_values(self, str currency, object time_frame, object from_timestamp=*, object to_timestamp=*)
    cpdef object get_historical_value(self, object timestamp)

    cdef void _add_historical_portfolio_value(self, double timestamp, dict value_by_currency)
    cdef bint _is_historical_timestamp_relevant(self, double timestamp, object time_frame_seconds, object from_timestamp, object to_timestamp)
    cdef set _get_relevant_timestamps(self, double timestamp, object currencies, list time_frames, bint force_update, bint include_past_data)
    cdef bint _should_update_timestamp(self, object currencies, object time_frame_allowed_window_start, bint force_update)
    cdef object _get_value_in_currency(self, object historical_value, str currency)
    cdef object _convert_historical_value(self, object historical_value, str target_currency)
    cdef object _get_metadata(self)
    cdef str _get_portfolio_type_suffix(self)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_commons.enums import TimeFrames, PriceIndexes
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc, \
    ExchangeConstantsOrderBookInfoColumns as Ecobic, ExchangeConstantsOrderColumns as Ecoc, \
    ExchangeConstantsTickersColumns as Ectc
from tests_additional.real_exchanges.real_exchange_tester import RealExchangeTester
# required to catch async loop context exceptions
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestHitBtcRealExchangeTester(RealExchangeTester):
    EXCHANGE_NAME = "hitbtc"
    SYMBOL = "BTC/USDT"
    SYMBOL_2 = "ETH/BTC"

    async def test_time_frames(self):
        time_frames = await self.time_frames()
        assert all(time_frame in time_frames for time_frame in (
            TimeFrames.ONE_MINUTE.value,
            TimeFrames.THREE_MINUTES.value,
            TimeFrames.FIVE_MINUTES.value,
            TimeFrames.FIFTEEN_MINUTES.value,
            TimeFrames.THIRTY_MINUTES.value,
            TimeFrames.ONE_HOUR.value,
            TimeFrames.FOUR_HOURS.value,
            TimeFrames.ONE_DAY.value,
            TimeFrames.ONE_WEEK.value,
            TimeFrames.ONE_MONTH.value
        ))

    async def test_get_market_status(self):
        for market_status in await self.get_market_statuses():
            assert market_status
            assert market_status[Ecmsc.SYMBOL.value] in (self.SYMBOL, self.SYMBOL_2)
            assert market_status[Ecmsc.PRECISION.value]
            assert 1e-08 <= market_status[Ecmsc.PRECISION.value][
                Ecmsc.PRECISION_AMOUNT.value] < 1  # to be fixed in hitbtc tentacle
            assert 1e-08 <= market_status[Ecmsc.PRECISION.value][
                Ecmsc.PRECISION_PRICE.value] < 1  # to be fixed in hitbtc tentacle
            assert all(elem in market_status[Ecmsc.LIMITS.value]
                       for elem in (Ecmsc.LIMITS_AMOUNT.value,
                                    Ecmsc.LIMITS_PRICE.value,
                                    Ecmsc.LIMITS_COST.value))

    async def test_get_symbol_prices(self):
        # without limit
        symbol_prices = await self.get_symbol_prices(sort='DESC')
        assert len(symbol_prices) == 100
        # check candles order (oldest first)
        self.ensure_elements_order(symbol_prices, PriceIndexes.IND_PRICE_TIME.value)
        # check last candle is the current candle
        assert symbol_prices[-1][PriceIndexes.IND_PRICE_TIME.value] >= self.get_time() - self.get_allowed_time_delta()

        # try with candles limit (used in candled updater)
        symbol_prices = await self.get_symbol_prices(limit=200, sort='DESC')  # to be fixed in hitbtc tentacle
        assert len(symbol_prices) == 200
        # check candles order (oldest first)
        self.ensure_elements_order(symbol_prices, PriceIndexes.IND_PRICE_TIME.value)
        # check last candle is the current candle
        assert symbol_prices[-1][PriceIndexes.IND_PRICE_TIME.value] >= self.get_time() - self.get_allowed_time_delta()

    async def test_get_kline_price(self):
        kline_price = await self.get_kline_price(sort='DESC')  # to be fixed in hitbtc tentacle
        assert len(kline_price) == 1
        assert len(kline_price[0]) == 6
        kline_start_time = kline_price[0][PriceIndexes.IND_PRICE_TIME.value]
        # assert kline is the current candle
        assert kline_start_time >= self.get_time() - self.get_allowed_time_delta()

    async def test_get_order_book(self):
        order_book = await self.get_order_book()
        assert len(order_book[Ecobic.ASKS.value]) == 5
        assert len(order_book[Ecobic.ASKS.value][0]) == 2
        assert len(order_book[Ecobic.BIDS.value]) == 5
        assert len(order_book[Ecobic.BIDS.value][0]) == 2

    async def test_get_recent_trades(self):
        recent_trades = await self.get_recent_trades()
        assert len(recent_trades) == 50
        # check trades order (oldest first)
        self.ensure_elements_order(recent_trades, Ecoc.TIMESTAMP.value)

    async def test_get_price_ticker(self):
        ticker = await self.get_price_ticker()
        self._check_ticker(ticker, self.SYMBOL, check_content=True)

    async def test_get_all_currencies_price_ticker(self):
        tickers = await self.get_all_currencies_price_ticker()
        for symbol, ticker in tickers.items():
            self._check_ticker(ticker, symbol)

    @staticmethod
    def _check_ticker(ticker, symbol, check_content=False):
        assert ticker[Ectc.SYMBOL.value] == symbol
        assert all(key in ticker for key in (
            Ectc.HIGH.value,
            Ectc.LOW.value,
            Ectc.BID.value,
            Ectc.BID_VOLUME.value,
            Ectc.ASK.value,
            Ectc.ASK_VOLUME.value,
            Ectc.OPEN.value,
            Ectc.CLOSE.value,
            Ectc.LAST.value,
            Ectc.PREVIOUS_CLOSE.value
        ))
        if check_content:
            assert ticker[Ectc.HIGH.value]
            assert ticker[Ectc.LOW.value]
            assert ticker[Ectc.BID.value]
            assert ticker[Ectc.BID_VOLUME.value] is None
            assert ticker[Ectc.ASK.value]
            assert ticker[Ectc.ASK_VOLUME.value] is None
            assert ticker[Ectc.OPEN.value]
            assert ticker[Ectc.CLOSE.value]
            assert ticker[Ectc.LAST.value]
            assert ticker[Ectc.PREVIOUS_CLOSE.value] is None
            assert ticker[Ectc.BASE_VOLUME.value]
            assert ticker[Ectc.TIMESTAMP.value]
            RealExchangeTester.check_ticker_typing(ticker)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.util as util
cimport octobot_trading.exchanges.config.backtesting_exchange_config as backtesting_exchange_config_import

cdef class ExchangeConfig(util.Initializable):
    cdef object _logger

    cdef public dict traded_cryptocurrencies
    cdef public dict config

    cdef public list traded_symbol_pairs
    cdef public list all_config_symbol_pairs
    cdef public list watched_pairs
    cdef public list available_required_time_frames
    cdef public list traded_time_frames
    cdef public list real_time_time_frames
    cdef public int required_historical_candles_count

    cdef public object exchange_manager

    cpdef backtesting_exchange_config_import.BacktestingExchangeConfig backtesting_exchange_config

    cpdef void set_config_time_frame(self)
    cpdef void set_config_traded_pairs(self)
    cpdef void set_historical_settings(self)
    cpdef object get_shortest_time_frame(self)
    cpdef void init_backtesting_exchange_config(self)

    @staticmethod
    cdef str _is_tradable_with_cryptocurrency(str symbol, str cryptocurrency)

    cdef void _set_config_time_frame(self)
    cdef void _set_config_traded_pairs(self)
    cdef set _set_config_traded_pair(self, str cryptocurrency, set traded_symbol_pairs_set, set existing_pairs)
    cdef void _populate_non_wildcard_pairs(self, str cryptocurrency, set existing_pairs, bint is_enabled)
    # return object to forward exceptions
    cdef object _populate_wildcard_pairs(self, str cryptocurrency, set existing_pairs, bint is_enabled)
    cdef list _add_tradable_symbols_from_config(self, str cryptocurrency, list filtered_symbols)
    cdef bint _is_valid_symbol(self, str symbol, list filtered_symbols)
    cdef list _create_wildcard_symbol_list(self, str cryptocurrency)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal
import contextlib

import octobot_commons.symbol_util as symbol_util
import octobot_commons.constants as commons_constants
import octobot_commons.logging as logging
import octobot_commons.timestamp_util as timestamp_util
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.exchanges.util.exchange_market_status_fixer as exchange_market_status_fixer
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc


LOGGER_NAME = "order_util"


def is_valid(element, key):
    """
    Checks is the element is valid with the market status fixer
    :param element:
    :param key:
    :return:
    """
    return key in element and exchange_market_status_fixer.is_ms_valid(element[key])


def get_min_max_amounts(symbol_market, default_value=None):
    """
    Returns the min and max quantity, cost and price according to the specified market
    :param symbol_market:
    :param default_value:
    :return:
    """
    min_quantity = max_quantity = min_cost = max_cost = min_price = max_price = default_value
    if Ecmsc.LIMITS.value in symbol_market:
        symbol_market_limits = symbol_market[Ecmsc.LIMITS.value]

        if Ecmsc.LIMITS_AMOUNT.value in symbol_market_limits:
            limit_amount = symbol_market_limits[Ecmsc.LIMITS_AMOUNT.value]
            if is_valid(limit_amount, Ecmsc.LIMITS_AMOUNT_MIN.value) \
                    or is_valid(limit_amount, Ecmsc.LIMITS_AMOUNT_MAX.value):
                min_quantity = limit_amount.get(Ecmsc.LIMITS_AMOUNT_MIN.value, default_value)
                max_quantity = limit_amount.get(Ecmsc.LIMITS_AMOUNT_MAX.value, default_value)

        # case 2: use cost and price
        if Ecmsc.LIMITS_COST.value in symbol_market_limits:
            limit_cost = symbol_market_limits[Ecmsc.LIMITS_COST.value]
            if is_valid(limit_cost, Ecmsc.LIMITS_COST_MIN.value) \
                    or is_valid(limit_cost, Ecmsc.LIMITS_COST_MAX.value):
                min_cost = limit_cost.get(Ecmsc.LIMITS_COST_MIN.value, default_value)
                max_cost = limit_cost.get(Ecmsc.LIMITS_COST_MAX.value, default_value)

        # case 2: use quantity and price
        if Ecmsc.LIMITS_PRICE.value in symbol_market_limits:
            limit_price = symbol_market_limits[Ecmsc.LIMITS_PRICE.value]
            if is_valid(limit_price, Ecmsc.LIMITS_PRICE_MIN.value) \
                    or is_valid(limit_price, Ecmsc.LIMITS_PRICE_MAX.value):
                min_price = limit_price.get(Ecmsc.LIMITS_PRICE_MIN.value, default_value)
                max_price = limit_price.get(Ecmsc.LIMITS_PRICE_MAX.value, default_value)

    return min_quantity, max_quantity, min_cost, max_cost, min_price, max_price


def check_cost(total_order_price, min_cost):
    """
    Checks and adapts the quantity and price of the order to ensure it's exchange compliant:
    - are the quantity and price of the order compliant with the exchange's number of digits requirement
        => otherwise quantity will be truncated accordingly
    - is the quantity valid
    - are the order total price and quantity superior or equal to the exchange's minimum order requirement
        => otherwise order is impossible => returns empty list
    - if total cost data are unavailable:
    - is the price of the currency compliant with the exchange's price interval for this currency
        => otherwise order is impossible => returns empty list
    - are the order total price and quantity inferior or equal to the exchange's maximum order requirement
        => otherwise order is impossible as is => split order into smaller ones and returns the list
    => returns the quantity and price list of possible order(s)
    - if exchange symbol data are not enough
        => try fixing exchange data using ExchangeMarketStatusFixer are start again (once only)
    """
    if total_order_price < min_cost:
        if min_cost is None:
            logging.get_logger(LOGGER_NAME).error("Invalid min_cost from exchange")
        return False
    return True


async def get_up_to_date_price(exchange_manager, symbol: str, timeout: int = None, base_error: str = None):
    exchange_time = exchange_manager.exchange.get_exchange_current_time()
    base_error = base_error or f"Can't get the necessary price data to create a new order on the " \
                               f"{timestamp_util.convert_timestamp_to_datetime(exchange_time)} " \
                               f"(timestamp: {exchange_time}):"
    try:
        mark_price = await exchange_manager.exchange_symbols_data.get_exchange_symbol_data(symbol) \
            .prices_manager.get_mark_price(timeout=timeout)
    except asyncio.TimeoutError:
        raise asyncio.TimeoutError(f"{base_error} mark price is not available")
    except errors.UnreachableExchange as e:
        raise errors.UnreachableExchange(f"{base_error} exchange is unreachable") from e
    return decimal.Decimal(str(mark_price))


async def get_pre_order_data(exchange_manager, symbol: str, timeout: int = None,
                             portfolio_type=commons_constants.PORTFOLIO_AVAILABLE):
    mark_price = await get_up_to_date_price(exchange_manager, symbol, timeout=timeout)
    symbol_market = exchange_manager.exchange.get_market_status(symbol, with_fixer=False)

    currency, market = symbol_util.split_symbol(symbol)
    portfolio = exchange_manager.exchange_personal_data.portfolio_manager.portfolio
    currency_available = portfolio.get_currency_portfolio(currency).available \
        if portfolio_type == commons_constants.PORTFOLIO_AVAILABLE else portfolio.get_currency_portfolio(currency).total
    market_available = portfolio.get_currency_portfolio(market).available \
        if portfolio_type == commons_constants.PORTFOLIO_AVAILABLE else portfolio.get_currency_portfolio(market).total

    if exchange_manager.is_future:
        pair_future_contract = exchange_manager.exchange.get_pair_future_contract(symbol)
        if pair_future_contract.is_inverse_contract():
            currency_available *= pair_future_contract.current_leverage
            market_quantity = market_available = currency_available
        else:
            market_available *= pair_future_contract.current_leverage / mark_price
            market_quantity = currency_available = market_available
    elif exchange_manager.is_margin:
        market_quantity = constants.ZERO  # TODO
    else:
        market_quantity = market_available / mark_price if mark_price else constants.ZERO
    return currency_available, market_available, market_quantity, mark_price, symbol_market


def get_futures_max_order_size(exchange_manager, symbol, side, current_price, reduce_only,
                               current_symbol_holding, market_quantity):
    # use position margin when trading futures and reducing the position
    current_position = exchange_manager.exchange_personal_data.positions_manager.get_symbol_position(
        symbol,
        enums.PositionSide.BOTH
    )
    # ensure max position order size is taken into account
    new_position_side = current_position.side
    if new_position_side is enums.PositionSide.UNKNOWN:
        new_position_side = enums.PositionSide.LONG if side is enums.TradeOrderSide.BUY \
            else enums.PositionSide.SHORT

    contract_current_symbol_holding = current_symbol_holding
    contract_market_quantity = market_quantity
    if current_position.symbol_contract.is_inverse_contract():
        # use USD (in BTC/USD) for order sizes on inverse, convert BTC values into USD ones
        contract_current_symbol_holding = current_symbol_holding * current_price
        contract_market_quantity = market_quantity * current_price
    if side is enums.TradeOrderSide.SELL and current_position.is_long():
        # can also sell the position size in long
        contract_current_symbol_holding = current_position.size if reduce_only \
            else contract_current_symbol_holding + current_position.size
    elif side is enums.TradeOrderSide.BUY and current_position.is_short():
        # can also buy the position size in short
        contract_market_quantity = abs(current_position.size) if reduce_only \
            else contract_market_quantity + abs(current_position.size)
    if (new_position_side is enums.PositionSide.LONG and side is enums.TradeOrderSide.BUY) \
            or (new_position_side is enums.PositionSide.SHORT and
                side is enums.TradeOrderSide.SELL):
        quantity = market_quantity if current_position.symbol_contract.is_inverse_contract() \
            else market_quantity * current_price
        unleveraged_quantity = quantity / current_position.symbol_contract.current_leverage
        max_position_increased_order_quantity = get_max_order_quantity_for_price(
            current_position, unleveraged_quantity, current_price, new_position_side, symbol
        )
        # increasing position: always use the same currency
        return max_position_increased_order_quantity, True
    return contract_market_quantity if side is enums.TradeOrderSide.BUY else contract_current_symbol_holding, False


def get_max_order_quantity_for_price(position, available_quantity, price, side, symbol):
    """
    Returns the maximum order quantity in market or currency for given total usable funds, price and side.
    This amount is not the total usable funds as it also requires to keep the position's open order fees
    as well as the potential position liquidation fees in portfolio. Those fees are computed by
    get_two_way_taker_fee_for_quantity_and_price
    :param position: the position to compute quantity for
    :param available_quantity: the maximum amount of currency/market to allocate to the position (without leverage)
    :param price: the target entry price of the position
    :param side: the side of the position
    :param side: the symbol of the position
    :return: the computed leveraged maximum entry quantity
    """
    # use position.symbol_contract.current_leverage as quantity to simulate a 1 unit quantity (x leverage)
    two_way_fees = position.get_two_way_taker_fee_for_quantity_and_price(position.symbol_contract.current_leverage,
                                                                         price, side, symbol)
    if position.symbol_contract.is_inverse_contract():
        # Returns the maximum order quantity in market.
        return position.symbol_contract.current_leverage * available_quantity / \
            (two_way_fees + constants.ONE / price)
    # Returns the maximum order quantity in currency.
    return position.symbol_contract.current_leverage * available_quantity / \
        (two_way_fees + price)


def total_fees_from_order_dict(order_dict, currency):
    return get_fees_for_currency(order_dict[enums.ExchangeConstantsOrderColumns.FEE.value], currency)


def get_fees_for_currency(fee, currency):
    if fee and fee[enums.FeePropertyColumns.CURRENCY.value] == currency:
        return decimal.Decimal(str(fee[enums.FeePropertyColumns.COST.value]))
    return constants.ZERO


def parse_raw_fees(raw_fees):
    fees = raw_fees
    if fees and enums.ExchangeConstantsOrderColumns.COST.value in fees:
        raw_fees[enums.ExchangeConstantsOrderColumns.COST.value] = \
            decimal.Decimal(str(raw_fees[enums.ExchangeConstantsOrderColumns.COST.value]))
    return fees


def parse_order_status(raw_order):
    try:
        return enums.OrderStatus(raw_order[enums.ExchangeConstantsOrderColumns.STATUS.value])
    except KeyError:
        return KeyError("Could not parse new order status")


def parse_is_cancelled(raw_order):
    return parse_order_status(raw_order) in {enums.OrderStatus.CANCELED, enums.OrderStatus.CLOSED}


def get_pnl_transaction_source_from_order(order):
    if order.order_type in [enums.TraderOrderType.SELL_MARKET, enums.TraderOrderType.BUY_MARKET,
                            enums.TraderOrderType.TAKE_PROFIT]:
        return enums.PNLTransactionSource.MARKET_ORDER
    if order.order_type in [enums.TraderOrderType.SELL_LIMIT, enums.TraderOrderType.BUY_LIMIT,
                            enums.TraderOrderType.TAKE_PROFIT_LIMIT]:
        return enums.PNLTransactionSource.LIMIT_ORDER
    if is_stop_order(order.order_type):
        return enums.PNLTransactionSource.STOP_ORDER
    return enums.PNLTransactionSource.UNKNOWN


def is_stop_order(order_type):
    return order_type in [enums.TraderOrderType.STOP_LOSS, enums.TraderOrderType.STOP_LOSS_LIMIT,
                          enums.TraderOrderType.TRAILING_STOP, enums.TraderOrderType.TRAILING_STOP_LIMIT]


async def create_as_chained_order(order):
    order.is_waiting_for_chained_trigger = False
    if not order.trader.simulate and order.has_been_bundled:
        # exchange should have created it already, it is either already fetched or
        # will automatically be fetched at the next update
        # warning: not handling instantly filled bundled orders as there is no easy way to do this
        # TODO: figure out instantly filled bundled orders
        if not await _apply_pending_order_on_existing_orders(order):
            # register it as pending creation order for it to be found and update when fetched
            order.exchange_manager.exchange_personal_data.orders_manager.register_pending_creation_order(order)
    else:
        # set created now to consider creation failures as created as well (the caller can always retry later on)
        order.status = enums.OrderStatus.OPEN
        # set uninitialized to allow second initialization from create_order
        order.is_initialized = False
        await order.trader.create_order(
            order,
            loaded=False,
            params=order.exchange_creation_params,
            **order.trader_creation_kwargs
        )


def is_associated_pending_order(pending_order, created_order):
    return created_order.symbol == pending_order.symbol and \
           created_order.origin_quantity == pending_order.origin_quantity and \
           created_order.origin_price == pending_order.origin_price and \
           created_order.__class__ is pending_order.__class__ and \
           created_order.trader is pending_order.trader


async def apply_pending_order_from_created_order(pending_order, created_order, to_be_initialized):
    await pending_order.update_from_order(created_order)
    if to_be_initialized:
        pending_order.is_initialized = False
    logging.get_logger(LOGGER_NAME).debug(f"Updated pending order: {pending_order} using {created_order}")


async def _apply_pending_order_on_existing_orders(pending_order):
    for created_order in pending_order.exchange_manager.exchange_personal_data.orders_manager.get_open_orders(
            symbol=pending_order.symbol):
        if is_associated_pending_order(pending_order, created_order) and created_order.order_group is None:
            await apply_pending_order_from_created_order(pending_order, created_order, False)
            pending_order.exchange_manager.exchange_personal_data.orders_manager.replace_order(created_order.order_id,
                                                                                               pending_order)
            created_order.clear()
            return True
    return False


@contextlib.asynccontextmanager
async def ensure_orders_relevancy(order=None, position=None):
    exchange_manager = order.exchange_manager if position is None else position.exchange_manager
    # part used in futures trading only
    if exchange_manager.exchange_personal_data.positions_manager.positions:
        position = position or exchange_manager.exchange_personal_data.positions_manager.get_order_position(order)
        pre_update_position_side = position.side
        is_pre_update_position_idle = position.is_idle()
        yield
        if not is_pre_update_position_idle and \
           (position.side != pre_update_position_side or position.is_idle()):
            # when position side is changing (from a non-idle position) or is going back to idle,
            # then associated reduce only orders must be closed
            await _cancel_reduce_only_orders_on_position_reset(exchange_manager, position.symbol)
    else:
        # as a context manager, yield is mandatory
        yield


async def _cancel_reduce_only_orders_on_position_reset(exchange_manager, symbol):
    for order in list(exchange_manager.exchange_personal_data.orders_manager.get_open_orders(symbol)):
        # reduce only order are automatically cancelled on exchanges, only cancel simulated orders
        if (exchange_manager.is_trader_simulated or order.is_self_managed()) \
                and order.is_open() and order.reduce_only:
            await order.trader.cancel_order(order)
            if order.order_group:
                await order.order_group.on_cancel(order)


def get_order_quantity_currency(exchange_manager, symbol, side):
    try:
        base, quote = symbol_util.split_symbol(symbol)
    except ValueError:
        # symbol that can't be split
        return None
    if exchange_manager.is_future:
        position = exchange_manager.exchange_personal_data.positions_manager.get_symbol_position(
            symbol, side
        )
        return quote if position.symbol_contract.is_inverse_contract() else base
    # always base in spot
    return base



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_commons.symbol_util as symbol_util

import octobot_trading.util as util
import octobot_trading.constants as constants


class PortfolioProfitability:
    """
    PortfolioProfitability calculates the portfolio profitability
    by subtracting portfolio_current_value and portfolio_origin_value
    """

    def __init__(self, portfolio_manager):
        self.portfolio_manager = portfolio_manager
        self.value_manager = portfolio_manager.portfolio_value_holder
        self.logger = logging.get_logger(f"{self.__class__.__name__}["
                                         f"{self.portfolio_manager.exchange_manager.exchange_name}]")

        # profitability attributes
        self.profitability = constants.ZERO
        self.profitability_percent = constants.ZERO
        self.profitability_diff = constants.ZERO
        self.market_profitability_percent = constants.ZERO
        self.initial_portfolio_current_profitability = constants.ZERO

        # buffer of currencies excluding market only used currencies ex: conf = btc/usd, eth/btc, ltc/btc, here usd
        # is market only => not used to compute market average profitability
        self.traded_currencies_without_market_specific = set()

        # set of currencies that should be valuated because either present in config or as a reference market
        self.valuated_currencies = util.get_all_currencies(self.portfolio_manager.config, enabled_only=False)
        self.valuated_currencies.add(self.portfolio_manager.reference_market)

    def get_average_market_profitability(self):
        """
        Returns the % move average of all the watched cryptocurrencies between bot's start time and now
        :return: the average market profitability
        """
        self.portfolio_manager.portfolio_value_holder.get_current_crypto_currencies_values()
        return self._calculate_average_market_profitability()

    def update_profitability(self, force_recompute_origin_portfolio=False):
        """
        Get profitability calls get_currencies_prices to update required data
        Then calls get_portfolio_current_value to set the current value of portfolio_current_value attribute
        :return: True if changed else False
        """
        self._reset_before_profitability_calculation()
        try:
            self.portfolio_manager.handle_profitability_recalculation(force_recompute_origin_portfolio)
            self._update_profitability_calculation()
            return self.profitability_diff != constants.ZERO
        except KeyError as missing_data_exception:
            self.logger.warning(f"Missing ticker data to calculate profitability")
            self.logger.warning(f"Missing {missing_data_exception} ticker data to calculate profitability")
        except Exception as missing_data_exception:
            self.logger.exception(missing_data_exception, True, str(missing_data_exception))

    def _reset_before_profitability_calculation(self):
        """
        Prepare profitability calculation
        """
        self.profitability_diff = self.profitability_percent
        self.profitability = constants.ZERO
        self.profitability_percent = constants.ZERO
        self.market_profitability_percent = constants.ZERO
        self.initial_portfolio_current_profitability = constants.ZERO

    def _update_profitability_calculation(self):
        """
        Calculates the new portfolio profitability
        """
        initial_portfolio_current_value = self.value_manager.get_origin_portfolio_current_value()
        self.profitability = self.value_manager.portfolio_current_value - self.value_manager.portfolio_origin_value

        if self.value_manager.portfolio_origin_value > constants.ZERO:
            self.profitability_percent = (constants.ONE_HUNDRED * self.value_manager.portfolio_current_value /
                                          self.value_manager.portfolio_origin_value) - constants.ONE_HUNDRED
            self.initial_portfolio_current_profitability = (constants.ONE_HUNDRED * initial_portfolio_current_value /
                                                            self.value_manager.portfolio_origin_value) - \
                                                           constants.ONE_HUNDRED
        else:
            self.profitability_percent = constants.ZERO
        self._update_portfolio_delta()

    def _update_portfolio_delta(self):
        """
        Calculates difference between the current and the last portfolio
        """
        self.profitability_diff = self.profitability_percent - self.profitability_diff
        self.market_profitability_percent = self.get_average_market_profitability()

    def _calculate_average_market_profitability(self):
        """
        Calculate the average of all the watched cryptocurrencies between bot's start time and now
        :return: the calculation result
        """
        origin_values = [value / self.value_manager.origin_crypto_currencies_values[currency]
                         for currency, value
                         in self._only_symbol_currency_filter(self.value_manager.
                                                              current_crypto_currencies_values).items()
                         if self.value_manager.origin_crypto_currencies_values[currency] > constants.ZERO]

        return sum(origin_values) / len(origin_values) * constants.ONE_HUNDRED - constants.ONE_HUNDRED \
            if origin_values else constants.ZERO

    def _only_symbol_currency_filter(self, currency_dict):
        """
        Return the dict of traded currencies with their portfolio value
        :param currency_dict: the currency dictionary to be filtered
        :return: the currency portfolio value filtered
        """
        if not self.traded_currencies_without_market_specific:
            self._init_traded_currencies_without_market_specific()
        return {currency: v for currency, v in currency_dict.items()
                if currency in self.traded_currencies_without_market_specific}

    def _init_traded_currencies_without_market_specific(self):
        """
        Initialize traded currencies without market specific set
        Use exchange_config.all_config_symbol_pairs to take every config pair into account including disabled ones
        """
        self.traded_currencies_without_market_specific = set(
            symbol_util.split_symbol(pair)[0]
            for pair in self.portfolio_manager.exchange_manager.exchange_config.all_config_symbol_pairs
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_trading.exchange_data.exchange_symbols_data import ExchangeSymbolsData

# Import required fixtures
from tests import event_loop
from tests.exchanges import exchange_manager

pytestmark = pytest.mark.asyncio


@pytest.mark.usefixtures("event_loop", "exchange_manager")
@pytest.fixture(scope="function")
async def exchange_symbols_data(exchange_manager):
    return ExchangeSymbolsData(exchange_manager)


async def test_get_exchange_symbol_data(exchange_symbols_data):
    new_symbols_data = exchange_symbols_data.get_exchange_symbol_data("BTC/USDT")
    assert new_symbols_data is not None
    assert exchange_symbols_data.get_exchange_symbol_data("BTC/USDT") is new_symbols_data


async def test_get_exchange_symbol_data_without_creation(exchange_symbols_data):
    with pytest.raises(KeyError):
        exchange_symbols_data.get_exchange_symbol_data("BTC/USDT", allow_creation=False)
    exchange_symbols_data.get_exchange_symbol_data("ETH/USDT", allow_creation=True)
    with pytest.raises(KeyError):
        exchange_symbols_data.get_exchange_symbol_data("ETH/BTC", allow_creation=False)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.constants as channel_constants
import async_channel.enums as channel_enums
import async_channel.channels as channels
import async_channel.consumer as consumers
import async_channel.producer as producers

import octobot_trading.enums as enums


class RemoteTradingSignalChannelConsumer(consumers.Consumer):
    pass


class RemoteTradingSignalChannelInternalConsumer(consumers.InternalConsumer):
    pass


class RemoteTradingSignalChannelProducer(producers.Producer):
    async def send(self, signal, bot_id):
        for consumer in self.channel.get_filtered_consumers(strategy=signal.strategy,
                                                            exchange=signal.exchange,
                                                            symbol=signal.symbol,
                                                            version=signal.version,
                                                            bot_id=bot_id):
            await consumer.queue.put({
                enums.TradingSignalAttrs.STRATEGY.value: signal.strategy,
                enums.TradingSignalAttrs.EXCHANGE.value: signal.exchange,
                enums.TradingSignalAttrs.SYMBOL.value: signal.symbol,
                RemoteTradingSignalsChannel.VERSION_KEY: signal.version,
                RemoteTradingSignalsChannel.BOT_ID_KEY: bot_id,
                RemoteTradingSignalsChannel.SIGNAL_KEY: signal,
            })


class RemoteTradingSignalsChannel(channels.Channel):
    PRODUCER_CLASS = RemoteTradingSignalChannelProducer
    CONSUMER_CLASS = RemoteTradingSignalChannelConsumer
    DEFAULT_PRIORITY_LEVEL = channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value
    SIGNAL_KEY = "signal"
    BOT_ID_KEY = "bot_id"
    VERSION_KEY = "version"

    async def new_consumer(self,
                           callback: object = None,
                           consumer_instance: object = None,
                           size: int = 0,
                           priority_level: int = DEFAULT_PRIORITY_LEVEL,
                           strategy: str = channel_constants.CHANNEL_WILDCARD,
                           exchange=channel_constants.CHANNEL_WILDCARD,
                           symbol: str = channel_constants.CHANNEL_WILDCARD,
                           version: str = channel_constants.CHANNEL_WILDCARD,
                           bot_id: str = channel_constants.CHANNEL_WILDCARD):
        consumer = consumer_instance if consumer_instance else self.CONSUMER_CLASS(callback,
                                                                                   size=size,
                                                                                   priority_level=priority_level)
        await self._add_new_consumer_and_run(consumer,
                                             strategy=strategy,
                                             exchange=exchange,
                                             symbol=symbol,
                                             version=version,
                                             bot_id=bot_id)
        return consumer

    def get_filtered_consumers(self,
                               strategy=channel_constants.CHANNEL_WILDCARD,
                               exchange=channel_constants.CHANNEL_WILDCARD,
                               symbol=channel_constants.CHANNEL_WILDCARD,
                               version=channel_constants.CHANNEL_WILDCARD,
                               bot_id=channel_constants.CHANNEL_WILDCARD):
        return self.get_consumer_from_filters({
            enums.TradingSignalAttrs.STRATEGY.value: strategy,
            enums.TradingSignalAttrs.EXCHANGE.value: exchange,
            enums.TradingSignalAttrs.SYMBOL.value: symbol,
            RemoteTradingSignalsChannel.VERSION_KEY: version,
            RemoteTradingSignalsChannel.BOT_ID_KEY: bot_id,
        })

    async def _add_new_consumer_and_run(self, consumer,
                                        strategy=channel_constants.CHANNEL_WILDCARD,
                                        exchange=channel_constants.CHANNEL_WILDCARD,
                                        symbol=channel_constants.CHANNEL_WILDCARD,
                                        version=channel_constants.CHANNEL_WILDCARD,
                                        bot_id=channel_constants.CHANNEL_WILDCARD):
        consumer_filters: dict = {
            enums.TradingSignalAttrs.STRATEGY.value: strategy,
            enums.TradingSignalAttrs.EXCHANGE.value: exchange,
            enums.TradingSignalAttrs.SYMBOL.value: symbol,
            RemoteTradingSignalsChannel.VERSION_KEY: version,
            RemoteTradingSignalsChannel.BOT_ID_KEY: bot_id,
        }

        self.add_new_consumer(consumer, consumer_filters)
        await consumer.run()
        self.logger.debug(f"Consumer started for : "
                          f"[strategy={strategy},"
                          f" exchange={exchange},"
                          f" symbol={symbol},"
                          f" version={version}]"
                          f" bot_id={bot_id}]")

    async def subscribe_to_product_feed(self, feed_id):
        await self.producers[0].subscribe_to_product_feed(feed_id)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.util as util


cdef class PricesManager(util.Initializable):
    cdef object logger
    cdef object exchange_manager

    cdef public object valid_price_received_event
    cdef public object mark_price

    cdef public double mark_price_set_time
    cdef int price_validity

    cdef dict mark_price_from_sources

    cdef void _set_mark_price_value(self, object mark_price)
    cdef void _reset_prices(self)
    cdef void _ensure_price_validity(self)
    cdef bint _are_other_sources_valid(self, str mark_price_source)
    cdef int _compute_mark_price_validity_timeout(self)

    cpdef bint set_mark_price(self, object mark_price, str mark_price_source)

cpdef object calculate_mark_price_from_recent_trade_prices(list recent_trade_prices)



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_trading.errors as errors
import octobot_trading.exchange_data.recent_trades.channel.recent_trade as recent_trade_channel
import octobot_trading.constants as constants
import octobot_trading.enums as enums


class RecentTradeUpdater(recent_trade_channel.RecentTradeProducer):
    CHANNEL_NAME = constants.RECENT_TRADES_CHANNEL
    RECENT_TRADE_REFRESH_TIME = 5
    RECENT_TRADE_LIMIT = 20  # should be < to RecentTradesManager's MAX_TRADES_COUNT

    def __init__(self, channel):
        super().__init__(channel)
        self.refresh_time = RecentTradeUpdater.RECENT_TRADE_REFRESH_TIME

    async def init_recent_trades(self):
        try:
            for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                recent_trades = await self.channel.exchange_manager.exchange.\
                    get_recent_trades(pair, limit=self.RECENT_TRADE_LIMIT)
                if recent_trades:
                    await self.push(pair,
                                    list(map(self.channel.exchange_manager.exchange.clean_recent_trade,
                                             recent_trades)))
            await asyncio.sleep(self.refresh_time)
        except Exception as e:
            self.logger.exception(e, True, f"Fail to initialize recent trades : {e}")

    async def start(self):
        refresh_threshold = self.channel.exchange_manager.get_rest_pairs_refresh_threshold()
        if refresh_threshold is enums.RestExchangePairsRefreshMaxThresholds.MEDIUM:
            self.refresh_time = 9
        elif refresh_threshold is enums.RestExchangePairsRefreshMaxThresholds.SLOW:
            self.refresh_time = 15
        await self.init_recent_trades()
        # check if channel is not None to avoid attribute error if channel has been concurrently closed
        if self.channel is not None:
            if self.channel.is_paused:
                await self.pause()
            else:
                await self.start_update_loop()

    async def start_update_loop(self):
        while not self.should_stop and not self.channel.is_paused:
            try:
                for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                    recent_trades = await self.channel.exchange_manager.exchange.\
                        get_recent_trades(pair, limit=self.RECENT_TRADE_LIMIT)
                    if recent_trades:
                        try:
                            await self.push(pair,
                                            list(map(self.channel.exchange_manager.exchange.clean_recent_trade,
                                                     recent_trades)))
                        except TypeError:
                            pass
                await asyncio.sleep(self.refresh_time)
            except errors.FailedRequest as e:
                self.logger.warning(str(e))
            except errors.NotSupported:
                self.logger.warning(f"{self.channel.exchange_manager.exchange_name} is not supporting updates")
                await self.pause()
            except Exception as e:
                self.logger.exception(e, True, f"Fail to update recent trades : {e}")

    async def resume(self) -> None:
        await super().resume()
        if not self.is_running:
            await self.run()



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time
import contextlib
import importlib
import asyncio

import octobot_commons.logging as logging
import octobot_commons.databases as databases
import octobot_commons.enums as commons_enums
import octobot_commons.errors as commons_errors
import octobot_commons.constants as commons_constants
import octobot_commons.channels_name as channels_name
import async_channel.channels as channels
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.api as trading_api
import octobot_trading.modes.abstract_trading_mode as abstract_trading_mode
import octobot_trading.modes.channel as modes_channel
import octobot_trading.modes.script_keywords.context_management as context_management
import octobot_trading.modes.script_keywords.basic_keywords as basic_keywords
import octobot_trading.constants as trading_constants
import octobot_trading.enums as trading_enums
import octobot_trading.errors as errors
import octobot_trading.storage as storage
import octobot_trading.personal_data as personal_data
import octobot_backtesting.api as backtesting_api
import octobot_tentacles_manager.api as tentacles_manager_api


class AbstractScriptedTradingMode(abstract_trading_mode.AbstractTradingMode):
    TRADING_SCRIPT_MODULE = None
    BACKTESTING_SCRIPT_MODULE = None

    BACKTESTING_ID_BY_BOT_ID = {}
    INITIALIZED_DB_BY_BOT_ID = {}
    SAVED_RUN_METADATA_DB_BY_BOT_ID = {}
    INITIALIZED_TRADING_PAIR_BY_BOT_ID = {}

    def __init__(self, config, exchange_manager):
        super().__init__(config, exchange_manager)
        self.producer = AbstractScriptedTradingModeProducer
        self._live_script = None
        self._backtesting_script = None
        self.timestamp = time.time()
        self.script_name = None

        if exchange_manager:
            self.load_config()
            # add config folder to importable files to import the user script
            tentacles_manager_api.import_user_tentacles_config_folder(self.exchange_manager.tentacles_setup_config)

    def get_current_state(self) -> (str, float):
        return super().get_current_state()[0] if self.producers[0].state is None else self.producers[0].state.name, \
               "N/A"

    async def create_producers(self) -> list:
        mode_producer = self.producer(
            exchanges_channel.get_chan(trading_constants.MODE_CHANNEL, self.exchange_manager.id),
            self.config, self, self.exchange_manager)
        await mode_producer.run()
        return [mode_producer]

    async def create_consumers(self) -> list:
        try:
            await exchanges_channel.get_chan(channels_name.OctoBotTradingChannelsName.TRADES_CHANNEL.value,
                                             self.exchange_manager.id).new_consumer(
                self._trades_callback,
                symbol=self.symbol
            )
            import octobot_services.channel as services_channels
            user_commands_consumer = \
                await channels.get_chan(services_channels.UserCommandsChannel.get_name()).new_consumer(
                    self._user_commands_callback,
                    {"bot_id": self.bot_id, "subject": self.get_name()}
                )
            return [user_commands_consumer]
        except ImportError:
            self.logger.warning("Can't connect to services channels")
        except KeyError:
            return []
        return []

    async def _trades_callback(
            self,
            exchange: str,
            exchange_id: str,
            cryptocurrency: str,
            symbol: str,
            trade: dict,
            old_trade: bool,
    ):
        if trade[trading_enums.ExchangeConstantsOrderColumns.STATUS.value] != trading_enums.OrderStatus.CANCELED.value:
            db = storage.RunDatabasesProvider.instance().get_trades_db(self.bot_id, self.exchange_manager.exchange_name)
            await basic_keywords.store_trade(None, trade, exchange_manager=self.exchange_manager, writer=db)

    async def _user_commands_callback(self, bot_id, subject, action, data) -> None:
        self.logger.debug(f"Received {action} command.")
        if action == commons_enums.UserCommands.RELOAD_SCRIPT.value:
            await self.reload_script(live=True)
            await self.reload_script(live=False)
        elif action == commons_enums.UserCommands.CLEAR_PLOTTING_CACHE.value:
            await self.clear_plotting_cache()
        elif action == commons_enums.UserCommands.CLEAR_SIMULATED_ORDERS_CACHE.value:
            await self.clear_simulated_orders_cache()
        elif action == commons_enums.UserCommands.CLEAR_SIMULATED_TRADES_CACHE.value:
            await self.clear_simulated_trades_cache()
        elif action == commons_enums.UserCommands.CLEAR_SIMULATED_TRANSACTIONS_CACHE.value:
            await self.clear_simulated_transactions_cache()

    async def clear_simulated_orders_cache(self):
        await basic_keywords.clear_orders_cache(
            storage.RunDatabasesProvider.instance().get_orders_db(self.bot_id, self.exchange_manager.exchange_name)
        )

    async def clear_simulated_trades_cache(self):
        await basic_keywords.clear_trades_cache(
            storage.RunDatabasesProvider.instance().get_trades_db(self.bot_id, self.exchange_manager.exchange_name)
        )

    async def clear_simulated_transactions_cache(self):
        await basic_keywords.clear_trades_cache(
            storage.RunDatabasesProvider.instance().get_transactions_db(self.bot_id, self.exchange_manager.exchange_name)
        )

    async def clear_plotting_cache(self):
        await basic_keywords.clear_all_tables(
            storage.RunDatabasesProvider.instance().get_symbol_db(self.bot_id,
                                                                  self.exchange_manager.exchange_name,
                                                                  self.symbol)
        )

    @classmethod
    async def get_backtesting_plot(cls, exchange, symbol, backtesting_id, optimizer_id, optimization_campaign):
        ctx = context_management.Context.minimal(cls, logging.get_logger(cls.get_name()), exchange, symbol,
                                                 backtesting_id, optimizer_id, optimization_campaign)
        return await cls.get_script_from_module(cls.BACKTESTING_SCRIPT_MODULE)(ctx)

    @classmethod
    def get_is_symbol_wildcard(cls) -> bool:
        return False

    def get_script(self, live=True):
        return self._live_script if live else self._backtesting_script

    def register_script_module(self, script_module, live=True):
        if live:
            self.__class__.TRADING_SCRIPT_MODULE = script_module
            self._live_script = self.get_script_from_module(script_module)
        else:
            self.__class__.BACKTESTING_SCRIPT_MODULE = script_module
            self._backtesting_script = self.get_script_from_module(script_module)

    @staticmethod
    def get_script_from_module(module):
        return module.script

    async def reload_script(self, live=True):
        module = self.__class__.TRADING_SCRIPT_MODULE if live else self.__class__.BACKTESTING_SCRIPT_MODULE
        importlib.reload(module)
        self.register_script_module(module, live=live)
        # reload config
        self.load_config()
        if live:
            # todo cancel and restart live tasks
            await self.start_over_database()

    async def start_over_database(self):
        await self.clear_plotting_cache()
        symbol_db = storage.RunDatabasesProvider.instance().get_symbol_db(self.bot_id,
                                                                          self.exchange_manager.exchange_name,
                                                                          self.symbol)
        symbol_db.set_initialized_flags(False)
        for producer in self.producers:
            for time_frame, call_args in producer.last_call_by_timeframe.items():
                run_db = storage.RunDatabasesProvider.instance().get_run_db(self.bot_id)
                await basic_keywords.clear_user_inputs(run_db)
                await producer.init_user_inputs(False)
                run_db.set_initialized_flags(False, (time_frame, ))
                self.__class__.INITIALIZED_DB_BY_BOT_ID[self.bot_id] = False
                await databases.CacheManager().close_cache(commons_constants.UNPROVIDED_CACHE_IDENTIFIER,
                                                           reset_cache_db_ids=True)
                await producer.call_script(*call_args)

    @classmethod
    async def get_backtesting_id(cls, bot_id):
        try:
            return cls.BACKTESTING_ID_BY_BOT_ID[bot_id]
        except KeyError:
            raise RuntimeError(f"No backtesting id for bot_id: {bot_id}")

    def set_initialized_trading_pair_by_bot_id(self, symbol, time_frame, initialized):
        try:
            self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name][
                symbol][time_frame] = initialized
        except KeyError:
            if self.bot_id not in self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID:
                self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id] = {}
            if self.exchange_manager.exchange_name not in \
                    self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id]:
                self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name] = {}
            if symbol not in \
                    self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name]:
                self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][
                    self.exchange_manager.exchange_name][symbol] = {}
            if time_frame not in \
                    self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name][
                        symbol]:
                self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name][
                    symbol][time_frame] = initialized

    def get_initialized_trading_pair_by_bot_id(self, symbol, time_frame):
        return self.__class__.INITIALIZED_TRADING_PAIR_BY_BOT_ID[self.bot_id][self.exchange_manager.exchange_name][
                symbol][time_frame]


class AbstractScriptedTradingModeProducer(modes_channel.AbstractTradingModeProducer):

    async def get_backtesting_metadata(self, user_inputs) -> dict:
        """
        Override this method to get add addition metadata
        :return: the metadata dict related to this backtesting run
        """
        symbols = trading_api.get_trading_pairs(self.exchange_manager)
        profitability, profitability_percent, _, _, _ = trading_api.get_profitability_stats(self.exchange_manager)
        origin_portfolio = personal_data.portfolio_to_float(
            self.exchange_manager.exchange_personal_data.portfolio_manager.
            portfolio_value_holder.origin_portfolio.portfolio)
        end_portfolio = personal_data.portfolio_to_float(
            self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio)
        for portfolio in (origin_portfolio, end_portfolio):
            for values in portfolio.values():
                values.pop("available", None)
        if self.exchange_manager.is_future:
            for position in self.exchange_manager.exchange_personal_data.positions_manager.positions.values():
                end_portfolio[position.get_currency()]["position"] = float(position.quantity)
        time_frames = [tf.value
                       for tf in trading_api.get_exchange_available_required_time_frames(self.exchange_name,
                                                                                         self.exchange_manager.id)]
        formatted_user_inputs = {}
        for user_input in user_inputs:
            if not user_input["is_nested_config"]:
                try:
                    formatted_user_inputs[user_input["tentacle"]][user_input["name"]] = user_input["value"]
                except KeyError:
                    formatted_user_inputs[user_input["tentacle"]] = {
                        user_input["name"]: user_input["value"]
                    }
        leverage = 0
        if self.exchange_manager.is_future and hasattr(self.exchange_manager.exchange, "get_pair_future_contract"):
            leverage = float(self.exchange_manager.exchange.get_pair_future_contract(symbols[0]).current_leverage)
        trades = trading_api.get_trade_history(self.exchange_manager)
        entries = [
            trade
            for trade in trades
            if trade.status is trading_enums.OrderStatus.FILLED and trade.side is trading_enums.TradeOrderSide.BUY
        ]
        win_rate = round(float(trading_api.get_win_rate(self.exchange_manager) * 100), 3)
        wins = round(win_rate * len(entries) / 100)
        draw_down = round(float(trading_api.get_draw_down(self.exchange_manager)), 3)
        r_sq_end_balance \
            = await trading_api.get_coefficient_of_determination(self.exchange_manager,
                                                                 use_high_instead_of_end_balance=False)
        r_sq_max_balance = await trading_api.get_coefficient_of_determination(self.exchange_manager)

        return {
            commons_enums.BacktestingMetadata.OPTIMIZATION_CAMPAIGN.value:
                self.run_dbs_identifier.optimization_campaign_name,
            commons_enums.BacktestingMetadata.ID.value: self.run_dbs_identifier.backtesting_id,
            commons_enums.BacktestingMetadata.GAINS.value: round(float(profitability), 8),
            commons_enums.BacktestingMetadata.PERCENT_GAINS.value: round(float(profitability_percent), 3),
            commons_enums.BacktestingMetadata.END_PORTFOLIO.value: str(end_portfolio),
            commons_enums.BacktestingMetadata.START_PORTFOLIO.value: str(origin_portfolio),
            commons_enums.BacktestingMetadata.WIN_RATE.value: win_rate,
            commons_enums.BacktestingMetadata.DRAW_DOWN.value: draw_down or 0,
            commons_enums.BacktestingMetadata.COEFFICIENT_OF_DETERMINATION_MAX_BALANCE.value: r_sq_max_balance or 0,
            commons_enums.BacktestingMetadata.COEFFICIENT_OF_DETERMINATION_END_BALANCE.value: r_sq_end_balance or 0,
            commons_enums.BacktestingMetadata.SYMBOLS.value: symbols,
            commons_enums.BacktestingMetadata.TIME_FRAMES.value: time_frames,
            commons_enums.BacktestingMetadata.START_TIME.value: backtesting_api.get_backtesting_starting_time(
                self.exchange_manager.exchange.backtesting),
            commons_enums.BacktestingMetadata.END_TIME.value: backtesting_api.get_backtesting_ending_time(
                self.exchange_manager.exchange.backtesting),
            commons_enums.BacktestingMetadata.DURATION.value: round(backtesting_api.get_backtesting_duration(
                self.exchange_manager.exchange.backtesting), 3),
            commons_enums.BacktestingMetadata.ENTRIES.value: len(entries),
            commons_enums.BacktestingMetadata.WINS.value: wins,
            commons_enums.BacktestingMetadata.LOSES.value: len(entries) - wins,
            commons_enums.BacktestingMetadata.TRADES.value: len(trades),
            commons_enums.BacktestingMetadata.TIMESTAMP.value: self.trading_mode.timestamp,
            commons_enums.BacktestingMetadata.NAME.value: self.trading_mode.script_name,
            commons_enums.BacktestingMetadata.LEVERAGE.value: leverage,
            commons_enums.BacktestingMetadata.USER_INPUTS.value: formatted_user_inputs,
            commons_enums.BacktestingMetadata.BACKTESTING_FILES.value: trading_api.get_backtesting_data_files(
                self.exchange_manager),
            commons_enums.BacktestingMetadata.EXCHANGE.value: self.exchange_name
        }

    async def get_live_metadata(self):
        start_time = backtesting_api.get_backtesting_starting_time(self.exchange_manager.exchange.backtesting) \
            if trading_api.get_is_backtesting(self.exchange_manager) \
            else trading_api.get_exchange_current_time(self.exchange_manager)
        end_time = backtesting_api.get_backtesting_ending_time(self.exchange_manager.exchange.backtesting) \
            if trading_api.get_is_backtesting(self.exchange_manager) \
            else -1
        exchange_type = "spot"
        exchanges = [self.exchange_name]  # TODO multi exchange
        future_contracts_by_exchange = {}
        if self.exchange_manager.is_future and hasattr(self.exchange_manager.exchange, "pair_contracts"):
            exchange_type = "future"
            future_contracts_by_exchange = {
                self.exchange_name: {
                    symbol: {
                        "contract_type": contract.contract_type.value,
                        "position_mode": contract.position_mode.value,
                        "margin_type": contract.margin_type.value
                    } for symbol, contract in self.exchange_manager.exchange.pair_contracts.items()
                }
            }
        return {
            commons_enums.DBRows.REFERENCE_MARKET.value: trading_api.get_reference_market(self.config),
            commons_enums.DBRows.START_TIME.value: start_time,
            commons_enums.DBRows.END_TIME.value: end_time,
            commons_enums.DBRows.TRADING_TYPE.value: exchange_type,
            commons_enums.DBRows.EXCHANGES.value: exchanges,
            commons_enums.DBRows.FUTURE_CONTRACTS.value: future_contracts_by_exchange,
            commons_enums.DBRows.SYMBOLS.value: self.exchange_manager.exchange_config.traded_symbol_pairs,
        }

    def __init__(self, channel, config, trading_mode, exchange_manager):
        super().__init__(channel, config, trading_mode, exchange_manager)
        self.last_call_by_timeframe = {}
        self.traded_pair = trading_mode.symbol
        self.are_metadata_saved = False

    async def start(self) -> None:
        await super().start()

        self.run_dbs_identifier = storage.RunDatabasesProvider.instance().get_run_databases_identifier(
            self.trading_mode.bot_id
        )
        # register backtesting id
        self.trading_mode.__class__.BACKTESTING_ID_BY_BOT_ID[self.trading_mode.bot_id] = \
            self.run_dbs_identifier.backtesting_id
        await self.run_dbs_identifier.initialize(self.exchange_name)
        # refresh user inputs
        await self.init_user_inputs(True)
        if not self.exchange_manager.is_backtesting:
            asyncio.create_task(self._schedule_initialization_call())

    async def _schedule_initialization_call(self):
        # initialization call is a special call that does not trigger trades and allows the script
        # to be run at least once in order to initialize its configuration
        if self.exchange_manager.is_backtesting:
            # not necessary in backtesting
            return

        # fake an full candle call
        cryptocurrency, symbol, time_frame = self._get_initialization_call_args()
        # wait for symbol data to be initialized
        # TODO use trading ready event when done
        candle = await self._wait_for_symbol_init(symbol, time_frame, 30)
        if candle is None:
            self.logger.error(f"Can't initialize trading script: {symbol} {time_frame} candles are not fetched")
        await self.ohlcv_callback(self.exchange_name, self.exchange_manager.id, cryptocurrency, symbol, time_frame,
                                  candle, init_call=True)

    async def _wait_for_symbol_init(self, symbol, time_frame, timeout):
        # warning: should never be called in backtesting
        tf = commons_enums.TimeFrames(time_frame)
        t0 = time.time()
        while time.time() - t0 < timeout:
            try:
                if self.exchange_manager.is_future:
                    # wait for contracts to be loaded
                    _ = self.exchange_manager.exchange.pair_contracts[symbol]
                candles_manager = self.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
                    symbol,
                    allow_creation=False) \
                    .symbol_candles[tf]
                candle_data = candles_manager.get_candles(5)
                current_time = self.exchange_manager.exchange.get_exchange_current_time()
                time_frame_sec = commons_enums.TimeFramesMinutes[tf] * commons_constants.MINUTE_TO_SECONDS
                last_full_candle_time = current_time - current_time % time_frame_sec - time_frame_sec
                for candle in reversed(candle_data):
                    if candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value] == last_full_candle_time:
                        return candle
                # return the candle right before the last (last being in construction)
                return candle_data[-2]
            except KeyError:
                # no symbol data initialized, keep waiting
                await asyncio.sleep(0.2)
        return None

    def _get_initialization_call_args(self):
        currency = next(iter(self.exchange_manager.exchange_config.traded_cryptocurrencies))
        symbol = self.exchange_manager.exchange_config.traded_cryptocurrencies[currency][0]
        time_frame = self.exchange_manager.exchange_config.traded_time_frames[0]
        return currency, symbol, time_frame.value

    async def ohlcv_callback(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                             time_frame: str, candle: dict, init_call: bool = False):
        with self.trading_mode_trigger():
            # add a full candle to time to get the real time
            trigger_time = candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value] + \
                           commons_enums.TimeFramesMinutes[commons_enums.TimeFrames(time_frame)] * \
                           commons_constants.MINUTE_TO_SECONDS
            await self.call_script(self.matrix_id, cryptocurrency, symbol, time_frame,
                                   commons_enums.ActivationTopics.FULL_CANDLES.value,
                                   trigger_time,
                                   candle=candle,
                                   init_call=init_call)

    async def kline_callback(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                             time_frame, kline: dict):
        with self.trading_mode_trigger():
            await self.call_script(self.matrix_id, cryptocurrency, symbol, time_frame,
                                   commons_enums.ActivationTopics.IN_CONSTRUCTION_CANDLES.value,
                                   kline[commons_enums.PriceIndexes.IND_PRICE_TIME.value],
                                   kline=kline)

    async def set_final_eval(self, matrix_id: str, cryptocurrency: str, symbol: str, time_frame):
        await self.call_script(matrix_id, cryptocurrency, symbol, time_frame,
                               commons_enums.ActivationTopics.EVALUATION_CYCLE.value,
                               self._get_latest_eval_time(matrix_id, cryptocurrency, symbol, time_frame))

    def _get_latest_eval_time(self, matrix_id: str, cryptocurrency: str, symbol: str, time_frame):
        try:
            import octobot_evaluators.matrix as matrix
            import octobot_evaluators.enums as evaluators_enums
            return matrix.get_latest_eval_time(matrix_id,
                                               exchange_name=self.exchange_name,
                                               tentacle_type=evaluators_enums.EvaluatorMatrixTypes.SCRIPTED.value,
                                               cryptocurrency=cryptocurrency,
                                               symbol=symbol,
                                               time_frame=time_frame)
        except ImportError:
            self.logger.error("OctoBot-Evaluators is required for a matrix callback")
            return None

    async def call_script(self, matrix_id: str, cryptocurrency: str, symbol: str, time_frame: str,
                          trigger_source: str, trigger_cache_timestamp: float,
                          candle: dict = None, kline: dict = None, init_call: bool = False):
        context = self.get_context(matrix_id, cryptocurrency, symbol, time_frame, trigger_source,
                                   trigger_cache_timestamp, candle, kline, init_call)
        self.last_call_by_timeframe[time_frame] = \
            (matrix_id, cryptocurrency, symbol, time_frame, trigger_source, trigger_cache_timestamp, candle, kline, init_call)
        context.matrix_id = matrix_id
        context.cryptocurrency = cryptocurrency
        context.symbol = symbol
        context.time_frame = time_frame
        initialized = True
        run_data_writer = storage.RunDatabasesProvider.instance().get_run_db(self.trading_mode.bot_id)
        try:
            if not run_data_writer.are_data_initialized and not \
                    self.trading_mode.__class__.INITIALIZED_DB_BY_BOT_ID.get(self.trading_mode.bot_id, False):
                await self._reset_run_data(context)
            await self._pre_script_call(context)
            await self.trading_mode.get_script(live=True)(context)
        except errors.UnreachableExchange:
            raise
        except (commons_errors.MissingDataError, commons_errors.ExecutionAborted) as e:
            self.logger.debug(f"Script execution aborted: {e}")
            initialized = run_data_writer.are_data_initialized
        except Exception as e:
            self.logger.exception(e, True, f"Error when running script: {e}")
        finally:
            if not self.exchange_manager.is_backtesting:
                # update db after each run only in live mode
                for database in self.all_databases().values():
                    if database:
                        await database.flush()
                if context.has_cache(context.symbol, context.time_frame):
                    await context.get_cache().flush()
            run_data_writer.set_initialized_flags(initialized)
            storage.RunDatabasesProvider.instance().get_symbol_db(self.trading_mode.bot_id,
                                                                  self.exchange_name, symbol)\
                .set_initialized_flags(initialized, (time_frame,))

    def get_context(self, matrix_id, cryptocurrency, symbol, time_frame, trigger_source, trigger_cache_timestamp,
                    candle, kline, init_call=False):
        context = context_management.Context(
            self.trading_mode,
            self.exchange_manager,
            self.exchange_manager.trader,
            self.exchange_name,
            self.traded_pair,
            matrix_id,
            cryptocurrency,
            symbol,
            time_frame,
            self.logger,
            self.trading_mode.__class__,
            trigger_cache_timestamp,
            trigger_source,
            candle or kline,
            None,
            None,
        )
        context.enable_trading = not init_call
        return context

    async def _reset_run_data(self, context):
        run_data_writer = storage.RunDatabasesProvider.instance().get_run_db(self.trading_mode.bot_id)
        await basic_keywords.clear_run_data(run_data_writer)
        await basic_keywords.save_metadata(run_data_writer, await self.get_live_metadata())
        await basic_keywords.save_portfolio(run_data_writer, context)
        self.trading_mode.__class__.INITIALIZED_DB_BY_BOT_ID[self.trading_mode.bot_id] = True

    async def init_user_inputs(self, should_clear_inputs):
        if should_clear_inputs:
            await basic_keywords.clear_user_inputs(
                storage.RunDatabasesProvider.instance().get_run_db(self.trading_mode.bot_id)
            )
        await self._register_required_user_inputs(
            self.get_context(None, None, self.trading_mode.symbol, None, None, None, None, None, True))

    async def _pre_script_call(self, context):
        await basic_keywords.set_leverage(context, await basic_keywords.user_select_leverage(context))

    async def _register_required_user_inputs(self, context):
        if context.exchange_manager.is_future:
            await basic_keywords.user_select_leverage(context)

        # register activating topics user input
        activation_topic_values = [
            commons_enums.ActivationTopics.EVALUATION_CYCLE.value,
            commons_enums.ActivationTopics.FULL_CANDLES.value,
            commons_enums.ActivationTopics.IN_CONSTRUCTION_CANDLES.value
        ]
        await basic_keywords.get_activation_topics(
            context,
            commons_enums.ActivationTopics.EVALUATION_CYCLE.value,
            activation_topic_values
        )

    @contextlib.asynccontextmanager
    async def get_metadata_writer(self, with_lock):
        async with databases.DBWriter.database(self.run_dbs_identifier.get_backtesting_metadata_identifier(),
                                               with_lock=with_lock) as writer:
            yield writer

    async def _save_transactions(self):
        await basic_keywords.store_transactions(
            None,
            self.exchange_manager.exchange_personal_data.transactions_manager.transactions.values(),
            writer=storage.RunDatabasesProvider.instance().get_transactions_db(self.trading_mode.bot_id,
                                                                               self.exchange_name)
        )

    async def stop(self) -> None:
        """
        Stop trading mode channels subscriptions
        """
        if not self.are_metadata_saved and self.exchange_manager is not None and self.exchange_manager.is_backtesting:
            run_data_writer = storage.RunDatabasesProvider.instance().get_run_db(self.trading_mode.bot_id)
            await run_data_writer.flush()
            user_inputs = await basic_keywords.get_user_inputs(run_data_writer)
            await self._save_transactions()
            if not self.trading_mode.__class__.SAVED_RUN_METADATA_DB_BY_BOT_ID.get(self.trading_mode.bot_id, False):
                try:
                    self.trading_mode.__class__.SAVED_RUN_METADATA_DB_BY_BOT_ID[self.trading_mode.bot_id] = True
                    async with self.get_metadata_writer(with_lock=True) as writer:
                        await basic_keywords.save_metadata(writer, await self.get_backtesting_metadata(user_inputs))
                        self.are_metadata_saved = True
                except Exception:
                    self.trading_mode.__class__.SAVED_RUN_METADATA_DB_BY_BOT_ID[self.trading_mode.bot_id] = False
                    raise
        await super().stop()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchanges.util cimport exchange_market_status_fixer
from octobot_trading.exchanges.util.exchange_market_status_fixer cimport (
    ExchangeMarketStatusFixer,
    is_ms_valid,
)
from octobot_trading.exchanges.util cimport exchange_util
from octobot_trading.exchanges.util.exchange_util cimport (
    get_margin_exchange_class,
    get_future_exchange_class,
    get_spot_exchange_class,
    get_order_side,
    get_exchange_type,
)
from octobot_trading.exchanges.util cimport websockets_util
from octobot_trading.exchanges.util.websockets_util cimport (
    force_disable_web_socket,
    check_web_socket_config,
    search_websocket_class,
)

__all__ = [
    "ExchangeMarketStatusFixer",
    "is_ms_valid",
    "get_margin_exchange_class",
    "get_future_exchange_class",
    "get_spot_exchange_class",
    "get_order_side",
    "force_disable_web_socket",
    "check_web_socket_config",
    "search_websocket_class",
    "get_exchange_type",
]



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.util as util


cdef class AbstractExchange(util.Initializable):
    cdef public dict config

    cdef public object logger
    cdef public object current_account
    cdef public object trader
    cdef public object exchange_manager
    cdef public object connector

    cdef public set symbols
    cdef public set time_frames

    cdef public double allowed_time_lag

    cdef public str name
    cdef public bint is_unreachable

    cpdef str get_name(cls)

    # exchange requests
    cpdef dict get_market_status(self, str symbol, object price_example=*, bint with_fixer=*)
    cpdef dict get_trade_fee(self, str symbol, object order_type, object quantity, object price, str taker_or_maker)
    cpdef dict get_fees(self, str symbol)
    cpdef double get_uniform_timestamp(self, double timestamp)
    cpdef str get_pair_from_exchange(self, str pair)
    cpdef str get_exchange_pair(self, str pair)
    cpdef str get_pair_cryptocurrency(self, str pair)
    cpdef tuple get_split_pair_from_exchange(self, str pair)
    cpdef int get_rate_limit(self)
    cpdef dict get_default_balance(self)

    # exchange settings
    cpdef bint authenticated(self)
    cpdef int get_max_handled_pair_with_time_frame(self)

    # parsers
    cpdef dict parse_balance(self, dict balance)
    cpdef dict parse_trade(self, dict trade)
    cpdef dict parse_order(self, dict order)
    cpdef dict parse_ticker(self, dict ticker)
    cpdef dict parse_ohlcv(self, dict ohlcv)
    cpdef dict parse_order_book(self, dict order_book)
    cpdef dict parse_order_book_ticker(self, dict order_book_ticker)
    cpdef double parse_timestamp(self, dict data_dict, str timestamp_key, object default_value=*, bint ms=*)
    cpdef str parse_currency(self, str currency)
    cpdef str parse_order_id(self, dict order)
    cpdef object parse_status(self, str status)
    cpdef object parse_side(self, str side)
    cpdef object parse_account(self, str account)

    # cleaners
    cpdef dict clean_recent_trade(self, dict recent_trade)
    cpdef dict clean_trade(self, dict trade)
    cpdef dict clean_order(self, dict order)

    # uniformization
    cpdef double get_exchange_current_time(self)
    cpdef object uniformize_candles_if_necessary(self, object candle_or_candles)
    cpdef object get_uniformized_timestamp(self, object candle_or_candles)
    cpdef long long get_candle_since_timestamp(self, object time_frame, int count)

    cdef object _uniformize_candles_timestamps(self, list candles)
    cdef void _uniformize_candle_timestamps(self, list candle)

    # utils
    cpdef void log_order_creation_error(self, object error, object order_type, str symbol, object quantity,
                                        object price, object stop_price)
    cpdef void handle_token_error(self, object error)
    cpdef bint is_supported_order_type(self, object order_type)
    cpdef bint supports_bundled_order_on_order_creation(self, object base_order, object bundled_order_type)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.personal_data.orders.order as order_class

cdef class TrailingStopOrder(order_class.Order):
    cdef object trailing_stop_price_hit_event # object is asyncio.Event
    cdef object trailing_price_hit_event # object is asyncio.Event
    cdef object wait_for_stop_price_hit_event_task # object is asyncio.Event
    cdef object wait_for_price_hit_event_task # object is asyncio.Event
    cdef public object trailing_percent
    cdef public bint allow_instant_fill

    cdef void _create_hit_events(self, object price_events_manager,
                                 object new_price,
                                 double new_price_time)
    cdef object _calculate_stop_price(self, object new_price)
    cdef void _remove_events(self, object price_events_manager)
    cdef void _clear_event_and_tasks(self)
    cdef void _cancel_hit_tasks(self)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants


async def user_input(
    ctx,
    name,
    input_type,
    def_val,
    min_val=None,
    max_val=None,
    options=None,
    is_nested_config=None,
    nested_tentacle=None,
    show_in_summary=True,
    show_in_optimizer=True,
    order=None,
    flush_if_necessary=False
):
    """
    Set and return a user input value.
    Types are: int, float, boolean, options, multiple-options, text
    :return:
    """
    tentacle_type_str = "trading_mode" if hasattr(ctx.tentacle, "trading_config") else "evaluator"
    config = ctx.tentacle.trading_config if tentacle_type_str == "trading_mode" else ctx.tentacle.specific_config
    try:
        value = config[name.replace(" ", "_")]
    except KeyError:
        config[name.replace(" ", "_")] = def_val
        value = def_val
    await save_user_input(
        ctx,
        name,
        input_type,
        value,
        def_val,
        tentacle_type_str,
        min_val=min_val,
        max_val=max_val,
        options=options,
        is_nested_config=is_nested_config,
        nested_tentacle=nested_tentacle,
        show_in_summary=show_in_summary,
        show_in_optimizer=show_in_optimizer,
        order=order,
        flush_if_necessary=flush_if_necessary
    )
    return value


async def save_user_input(
    ctx,
    name,
    input_type,
    value,
    def_val,
    tentacle_type_str,
    min_val=None,
    max_val=None,
    options=None,
    is_nested_config=None,
    nested_tentacle=None,
    show_in_summary=True,
    show_in_optimizer=True,
    order=None,
    flush_if_necessary=False
):
    if is_nested_config is None:
        is_nested_config = ctx.is_nested_tentacle
    if not await ctx.run_data_writer.contains_row(
            commons_enums.DBTables.INPUTS.value,
            {
                "name": name,
                "tentacle": ctx.tentacle.get_name(),
                "nested_tentacle": nested_tentacle,
                "is_nested_config": is_nested_config
            }):
        await ctx.run_data_writer.log(
            commons_enums.DBTables.INPUTS.value,
            {
                "name": name,
                "input_type": input_type,
                "value": value,
                "def_val": def_val,
                "min_val": min_val,
                "max_val": max_val,
                "options": options,
                "tentacle_type": tentacle_type_str,
                "tentacle": ctx.tentacle.get_name(),
                "nested_tentacle": nested_tentacle,
                "is_nested_config": is_nested_config,
                "in_summary": show_in_summary,
                "in_optimizer": show_in_optimizer,
                "order": order
            }
        )
        if not ctx.exchange_manager.is_backtesting and (flush_if_necessary or ctx.run_data_writer.are_data_initialized):
            # in some cases, user inputs might be setup after the 1st trading mode cycle: flush
            # writer in live mode to ensure writing
            await ctx.run_data_writer.flush()


async def get_user_inputs(reader):
    return await reader.all(commons_enums.DBTables.INPUTS.value)


async def clear_user_inputs(writer):
    await writer.delete(commons_enums.DBTables.INPUTS.value, None)


async def get_activation_topics(context, default_value, options):
    return await user_input(
        context, commons_constants.CONFIG_ACTIVATION_TOPICS, "options",
        default_value,
        options=options,
        show_in_optimizer=False, show_in_summary=False, order=1000
    )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.signals.trading_signal as trading_signal
import octobot_trading.enums as enums
import octobot_commons.enums as commons_enums


def create_trading_signal(dict_signal):
    signal_value = dict_signal[commons_enums.CommunityFeedAttrs.VALUE.value]
    return trading_signal.TradingSignal(
        signal_value.get(enums.TradingSignalAttrs.STRATEGY.value),
        signal_value.get(enums.TradingSignalAttrs.EXCHANGE.value),
        signal_value.get(enums.TradingSignalAttrs.EXCHANGE_TYPE.value),
        signal_value.get(enums.TradingSignalAttrs.SYMBOL.value),
        signal_value.get(enums.TradingSignalAttrs.DESCRIPTION.value),
        signal_value.get(enums.TradingSignalAttrs.STATE.value),
        signal_value.get(enums.TradingSignalAttrs.ORDERS.value),
        identifier=dict_signal[commons_enums.CommunityFeedAttrs.ID.value],
        version=dict_signal[commons_enums.CommunityFeedAttrs.VERSION.value],
    )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.constants
import octobot_trading.exchanges as exchanges


def is_exchange_managed_by_websocket(exchange_manager, channel):
    """
    # TODO improve checker
    """
    return not exchange_manager.rest_only \
           and exchange_manager.has_websocket \
           and not exchange_manager.is_backtesting \
           and channel in octobot_trading.constants.WEBSOCKET_FEEDS_TO_TRADING_CHANNELS \
           and any([exchange_manager.exchange_web_socket.is_feed_available(feed)
                    for feed in octobot_trading.constants.WEBSOCKET_FEEDS_TO_TRADING_CHANNELS[channel]])


def is_websocket_feed_requiring_init(exchange_manager, channel):
    return any([exchange_manager.exchange_web_socket.is_feed_requiring_init(feed)
                for feed in octobot_trading.constants.WEBSOCKET_FEEDS_TO_TRADING_CHANNELS[channel]])


async def search_and_create_websocket(exchange_manager):
    socket_manager = exchanges.search_websocket_class(exchanges.WebSocketExchange, exchange_manager)
    if socket_manager is not None:
        await _create_websocket(exchange_manager, exchanges.WebSocketExchange.__name__, socket_manager)


async def _create_websocket(exchange_manager, websocket_class_name, socket_manager):
    try:
        exchange_manager.exchange_web_socket = socket_manager.get_websocket_client(exchange_manager.config,
                                                                                   exchange_manager)
        if exchange_manager.is_valid_account:
            await _init_websocket(exchange_manager)
            exchange_manager.logger.info(f"{socket_manager.get_name()} connected to "
                                         f"{exchange_manager.exchange.name.capitalize()}")
        else:
            exchange_manager.logger.error(f"Impossible to start websockets for "
                                          f"{exchange_manager.exchange.name.capitalize()}: "
                                          f"incompatible account. Your OctoBot will work normally but will be "
                                          f"limited to the {exchange_manager.exchange.name.capitalize()} REST API "
                                          f"which is slower and allows for less simultaneous traded pairs "
                                          f"because of the exchange's rate limit.")
            exchange_manager.exchange_web_socket = None
            exchange_manager.has_websocket = False
    except Exception as e:
        exchange_manager.logger.error(f"Fail to init websocket for {websocket_class_name} "
                                      f"({exchange_manager.exchange.name}): {e}")
        exchange_manager.exchange_web_socket = None
        exchange_manager.has_websocket = False
        raise e


async def _init_websocket(exchange_manager):
    await exchange_manager.exchange_web_socket.init_websocket(exchange_manager.exchange_config.traded_time_frames,
                                                              exchange_manager.exchange_config.traded_symbol_pairs,
                                                              exchange_manager.tentacles_setup_config)

    await exchange_manager.exchange_web_socket.start_sockets()

    exchange_manager.has_websocket = exchange_manager.exchange_web_socket.is_websocket_running



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.constants as constants
import octobot_trading.personal_data.orders.channel.orders_updater as orders_updater


class OrdersUpdaterSimulator(orders_updater.OrdersUpdater):
    async def start(self):
        self.channel.exchange_manager.exchange_personal_data.orders_manager.are_exchange_orders_initialized = True
        await exchange_channel.get_chan(constants.RECENT_TRADES_CHANNEL, self.channel.exchange_manager.id) \
            .new_consumer(self.ignore_recent_trades_update)

    async def ignore_recent_trades_update(self, exchange: str, exchange_id: str,
                                          cryptocurrency: str, symbol: str, recent_trades: list):
        """
        Used to subscribe at least one recent trades consumer during backtesting
        """



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums


class TradingSignal:
    def __init__(self, strategy, exchange, exchange_type, symbol, description, state, orders,
                 identifier=None, version=None):
        self.strategy = strategy
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.symbol = symbol
        self.description = description
        self.state = state
        self.orders = orders

        self.identifier = identifier
        self.version = version or self._get_version()

    def to_dict(self):
        return {
            enums.TradingSignalAttrs.STRATEGY.value: self.strategy,
            enums.TradingSignalAttrs.EXCHANGE.value: self.exchange,
            enums.TradingSignalAttrs.EXCHANGE_TYPE.value: self.exchange_type,
            enums.TradingSignalAttrs.SYMBOL.value: self.symbol,
            enums.TradingSignalAttrs.DESCRIPTION.value: self.description,
            enums.TradingSignalAttrs.STATE.value: self.state,
            enums.TradingSignalAttrs.ORDERS.value: self.orders,
        }

    def __str__(self):
        return f"{self.to_dict()}"

    def _get_version(self):
        try:
            import octobot.constants
            return octobot.constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION
        except ImportError:
            return "1.0.0"



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


from octobot_trading.exchanges.implementations.cryptofeed_websocket_exchange import (
    CryptofeedWebSocketExchange,
)
from octobot_trading.exchanges.implementations import spot_exchange_simulator
from octobot_trading.exchanges.implementations.spot_exchange_simulator import (
    SpotExchangeSimulator,
)
from octobot_trading.exchanges.implementations import future_exchange_simulator
from octobot_trading.exchanges.implementations.future_exchange_simulator import (
    FutureExchangeSimulator,
)
from octobot_trading.exchanges.implementations import future_ccxt_exchange
from octobot_trading.exchanges.implementations.future_ccxt_exchange import (
    FutureCCXTExchange,
)
from octobot_trading.exchanges.implementations import margin_exchange_simulator
from octobot_trading.exchanges.implementations.margin_exchange_simulator import (
    MarginExchangeSimulator,
)
from octobot_trading.exchanges.implementations import margin_ccxt_exchange
from octobot_trading.exchanges.implementations.margin_ccxt_exchange import (
    MarginCCXTExchange,
)
from octobot_trading.exchanges.implementations import spot_ccxt_exchange
from octobot_trading.exchanges.implementations.spot_ccxt_exchange import (
    SpotCCXTExchange,
)
from octobot_trading.exchanges.implementations import ccxt_websocket_exchange
from octobot_trading.exchanges.implementations.ccxt_websocket_exchange import (
    CCXTWebSocketExchange,
)
from octobot_trading.exchanges.implementations import default_spot_ccxt_exchange
from octobot_trading.exchanges.implementations.default_spot_ccxt_exchange import (
    DefaultCCXTSpotExchange,
)
from octobot_trading.exchanges.implementations import cryptofeed_websocket_exchange
from octobot_trading.exchanges.implementations.cryptofeed_websocket_exchange import (
    CryptofeedWebSocketExchange,
)

__all__ = [
    "SpotExchangeSimulator",
    "FutureExchangeSimulator",
    "MarginExchangeSimulator",
    "FutureCCXTExchange",
    "MarginCCXTExchange",
    "SpotCCXTExchange",
    "CCXTWebSocketExchange",
    "DefaultCCXTSpotExchange",
    "CryptofeedWebSocketExchange",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums


def create_position_instance_from_raw(trader, raw_position):
    """
    Creates a position instance from a raw position dictionary
    :param trader: the trader instance
    :param raw_position: the raw position dictionary
    :return: the created position
    """
    position_symbol_contract = trader.exchange_manager.exchange.get_pair_future_contract(
        raw_position.get(enums.ExchangeConstantsPositionColumns.SYMBOL.value))
    position = create_position_from_type(trader, position_symbol_contract)
    position.update_from_raw(raw_position)
    return position


def create_position_from_type(trader, symbol_contract):
    """
    Creates a position instance from a position type
    :param trader: the trader instance
    :param symbol_contract: the position symbol contract
    :return: the created position
    """
    return personal_data.TraderPositionTypeClasses[symbol_contract.contract_type](trader, symbol_contract)


def create_symbol_position(trader, symbol):
    """
    Creates a position for a specified symbol
    :param trader: the trader instance
    :param symbol: the position symbol
    :return: the created position
    """
    return create_position_instance_from_raw(trader, {
        enums.ExchangeConstantsPositionColumns.SYMBOL.value: symbol
    })



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import trading_backend

import octobot_commons.authentication as authentication
import octobot_trading.errors as errors
import octobot_trading.exchanges as exchanges


async def create_exchanges(exchange_manager):
    if exchange_manager.is_sandboxed:
        exchange_manager.logger.info(f"Using sandbox exchange for {exchange_manager.exchange_name}")

    if exchange_manager.is_backtesting:
        # simulated : create exchange simulator instance
        await create_simulated_exchange(exchange_manager)
        exchange_manager.load_constants()
    else:
        # real : create a rest or websocket exchange instance
        await create_real_exchange(exchange_manager)
        exchange_manager.load_constants()
        await initialize_real_exchange(exchange_manager)

    if not exchange_manager.exchange_only:
        # create exchange producers if necessary
        await exchanges.create_exchange_producers(exchange_manager)

    if exchange_manager.is_backtesting:
        await init_simulated_exchange(exchange_manager)

    exchange_manager.exchange_name = exchange_manager.exchange.name
    exchange_manager.is_ready = True


async def create_real_exchange(exchange_manager) -> None:
    """
    Create and initialize real REST exchange
    :param exchange_manager: the related exchange manager
    """
    await _create_rest_exchange(exchange_manager)

    try:
        await exchange_manager.exchange.initialize()
        _create_exchange_backend(exchange_manager)
        await _initialize_exchange_backend(exchange_manager)
    except errors.AuthenticationError:
        exchange_manager.logger.error("Authentication error, retrying without authentication...")
        exchange_manager.without_auth = True
        await create_real_exchange(exchange_manager)
        return


async def initialize_real_exchange(exchange_manager):
    if not exchange_manager.exchange_only:
        await exchanges.create_exchange_channels(exchange_manager)

    # create Websocket exchange if possible
    if not exchange_manager.rest_only:
        # search for websocket
        if exchanges.check_web_socket_config(exchange_manager.config, exchange_manager.exchange.name):
            await exchanges.search_and_create_websocket(exchange_manager)


def _create_exchange_backend(exchange_manager):
    try:
        exchange_manager.exchange_backend = trading_backend.exchange_factory.create_exchange_backend(
            exchange_manager.exchange
        )
    except Exception as e:
        exchange_manager.logger.exception(e, True, f"Error when creating exchange backend: {e}")


async def _initialize_exchange_backend(exchange_manager):
    if exchange_manager.exchange_backend is not None and exchange_manager.exchange.authenticated() \
            and not exchange_manager.is_trader_simulated:
        try:
            if await _is_supporting_octobot():
                exchange_manager.is_valid_account = True
                return True
            exchange_manager.is_valid_account, message = await exchange_manager.exchange_backend.is_valid_account()
            if not exchange_manager.is_valid_account:
                exchange_manager.logger.error(
                    f"Incompatible {exchange_manager.exchange.name.capitalize()} account to use websockets: {message}. "
                    f"OctoBot relies on exchanges profits sharing to remain 100% free, please create a "
                    f"new {exchange_manager.exchange.name.capitalize()} account to support the project. "
                    f"{exchanges.get_partners_explanation_message()}")
        except trading_backend.TimeSyncError as err:
            exchanges.log_time_sync_error(exchange_manager.logger, exchange_manager.exchange.name,
                                          err, "account details")
            exchange_manager.is_valid_account = False
        except Exception as e:
            exchange_manager.is_valid_account = False
            exchange_manager.logger.exception(e, True, f"Error when loading exchange account: {e}")
        finally:
            if exchange_manager.is_valid_account:
                exchange_manager.logger.info("On behalf of the OctoBot team, thank you for "
                                             "actively supporting the project !")


async def _is_supporting_octobot() -> bool:
    try:
        authenticator = authentication.Authenticator.instance()
        if not authenticator.is_initialized():
            initialization_timeout = 5
            await authenticator.await_initialization(initialization_timeout)
        if authenticator.supports.is_supporting():
            return True
    except asyncio.TimeoutError:
        pass
    return False


async def _create_rest_exchange(exchange_manager) -> None:
    """
    create REST based on ccxt exchange
    :param exchange_manager: the related exchange manager
    """
    if exchange_manager.is_future and not exchange_manager.is_spot_only:
        await _search_and_create_future_exchange(exchange_manager)
    elif exchange_manager.is_margin and not exchange_manager.is_spot_only:
        await _search_and_create_margin_exchange(exchange_manager)
    else:
        await search_and_create_spot_exchange(exchange_manager)

    if not exchange_manager.exchange:
        raise Exception("Can't create an exchange instance that match the exchange configuration")


async def create_simulated_exchange(exchange_manager):
    if exchange_manager.is_future and not exchange_manager.is_spot_only:
        exchange_manager.exchange = exchanges.FutureExchangeSimulator(exchange_manager.config, exchange_manager,
                                                                      exchange_manager.backtesting)
    elif exchange_manager.is_margin and not exchange_manager.is_spot_only:
        exchange_manager.exchange = exchanges.MarginExchangeSimulator(exchange_manager.config, exchange_manager,
                                                                      exchange_manager.backtesting)
    else:
        exchange_manager.exchange = exchanges.SpotExchangeSimulator(exchange_manager.config, exchange_manager,
                                                                    exchange_manager.backtesting)

    await exchange_manager.exchange.initialize()
    _initialize_simulator_time_frames(exchange_manager)
    exchange_manager.exchange_config.set_config_time_frame()
    exchange_manager.exchange_config.set_config_traded_pairs()
    await exchanges.create_exchange_channels(exchange_manager)


async def init_simulated_exchange(exchange_manager):
    await exchange_manager.exchange.create_backtesting_exchange_producers()


async def search_and_create_spot_exchange(exchange_manager) -> None:
    """
    Create a spot exchange if a SpotExchange matching class is found
    :param exchange_manager: the related exchange manager
    """
    spot_exchange_class = exchanges.get_spot_exchange_class(exchange_manager.exchange_name,
                                                            exchange_manager.tentacles_setup_config)
    if spot_exchange_class:
        exchange_manager.exchange = spot_exchange_class(config=exchange_manager.config,
                                                        exchange_manager=exchange_manager)


async def _search_and_create_margin_exchange(exchange_manager) -> None:
    """
    Create a margin exchange if a MarginExchange matching class is found
    :param exchange_manager: the related exchange manager
    """
    margin_exchange_class = exchanges.get_margin_exchange_class(exchange_manager.exchange_name,
                                                                exchange_manager.tentacles_setup_config)
    if margin_exchange_class:
        exchange_manager.exchange = margin_exchange_class(config=exchange_manager.config,
                                                          exchange_manager=exchange_manager)


async def _search_and_create_future_exchange(exchange_manager) -> None:
    """
    Create a future exchange if a FutureExchange matching class is found
    :param exchange_manager: the related exchange manager
    """
    future_exchange_class = exchanges.get_future_exchange_class(exchange_manager.exchange_name,
                                                                exchange_manager.tentacles_setup_config)
    if future_exchange_class:
        exchange_manager.exchange = future_exchange_class(config=exchange_manager.config,
                                                          exchange_manager=exchange_manager)


def _initialize_simulator_time_frames(exchange_manager):
    """
    Initialize simulator client time frames
    :param exchange_manager: the related exchange manager
    """
    exchange_manager.client_time_frames = exchange_manager.exchange.get_available_time_frames()



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_commons.singleton as singleton
cimport octobot_trading.exchanges.exchange_manager as trading_exchanges

cdef class ExchangeConfiguration(object):
    cdef public trading_exchanges.ExchangeManager exchange_manager

    cdef public str exchange_name
    cdef public str id
    cdef public str matrix_id

    cdef public dict symbols_by_crypto_currencies
    cdef public list symbols
    cdef public list real_time_time_frames
    cdef public list available_required_time_frames


cdef class Exchanges(singleton.Singleton):
    cdef public dict exchanges

    cpdef void add_exchange(self, trading_exchanges.ExchangeManager exchange_manager, str matrix_id)
    cpdef ExchangeConfiguration get_exchange(self, str exchange_name, str exchange_manager_id)
    cpdef dict get_exchanges(self, str exchange_name)
    cpdef list get_all_exchanges(self)
    cpdef list get_exchanges_list(self, str exchange_name)
    cpdef void del_exchange(self, str exchange_name, str exchange_manager_id, bint should_warn=*)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing

from octobot_commons import enums as common_enums

import octobot_trading.exchanges.connectors as exchange_connectors
import octobot_trading.exchanges.types as exchanges_types
import octobot_trading.personal_data as personal_data
from octobot_trading import enums as enums


class MarginCCXTExchange(exchanges_types.MarginExchange):
    def __init__(self, config, exchange_manager):
        super().__init__(config, exchange_manager)
        self.connector = exchange_connectors.CCXTExchange(
            config,
            exchange_manager,
            additional_ccxt_config=self.get_additional_connector_config()
        )

    async def initialize_impl(self):
        await self.connector.initialize()
        self.symbols = self.connector.symbols
        self.time_frames = self.connector.time_frames

    async def stop(self) -> None:
        await self.connector.stop()
        self.exchange_manager = None

    @classmethod
    def is_supporting_exchange(cls, exchange_candidate_name) -> bool:
        return exchange_connectors.CCXTExchange.is_supporting_exchange(exchange_candidate_name)

    def get_exchange_current_time(self):
        return self.connector.get_exchange_current_time()

    def get_uniform_timestamp(self, timestamp):
        return self.connector.get_uniform_timestamp(timestamp)

    def get_market_status(self, symbol, price_example=None, with_fixer=True):
        return self.connector.get_market_status(symbol, price_example=price_example, with_fixer=with_fixer)

    async def get_balance(self, **kwargs: dict):
        return await self.connector.get_balance(**kwargs)

    async def get_symbol_prices(self, symbol: str, time_frame: common_enums.TimeFrames, limit: int = None,
                                **kwargs: dict) -> typing.Optional[list]:
        return await self.connector.get_symbol_prices(symbol=symbol, time_frame=time_frame, limit=limit, **kwargs)

    async def get_kline_price(self, symbol: str, time_frame: common_enums.TimeFrames, **kwargs: dict) -> typing.Optional[list]:
        return await self.connector.get_kline_price(symbol=symbol, time_frame=time_frame, **kwargs)

    async def get_order_book(self, symbol: str, limit: int = 5, **kwargs: dict) -> typing.Optional[dict]:
        return await self.connector.get_order_book(symbol=symbol, limit=limit, **kwargs)

    async def get_recent_trades(self, symbol: str, limit: int = 50, **kwargs: dict) -> typing.Optional[list]:
        return await self.connector.get_recent_trades(symbol=symbol, limit=limit, **kwargs)

    async def get_price_ticker(self, symbol: str, **kwargs: dict) -> typing.Optional[dict]:
        return await self.connector.get_price_ticker(symbol=symbol, **kwargs)

    async def get_all_currencies_price_ticker(self, **kwargs: dict) -> typing.Optional[list]:
        return await self.connector.get_all_currencies_price_ticker(**kwargs)

    async def get_order(self, order_id: str, symbol: str = None, **kwargs: dict) -> dict:
        return await self.connector.get_order(symbol=symbol, order_id=order_id, **kwargs)

    async def get_all_orders(self, symbol: str = None, since: int = None, limit: int = None, **kwargs: dict) -> list:
        return await self.connector.get_all_orders(symbol=symbol, since=since, limit=limit, **kwargs)

    async def get_open_orders(self, symbol: str = None, since: int = None, limit: int = None, **kwargs: dict) -> list:
        return await self.connector.get_open_orders(symbol=symbol, since=since, limit=limit, **kwargs)

    async def get_closed_orders(self, symbol: str = None, since: int = None, limit: int = None, **kwargs: dict) -> list:
        return await self.connector.get_closed_orders(symbol=symbol, since=since, limit=limit, **kwargs)

    async def get_my_recent_trades(self, symbol: str = None, since: int = None, limit: int = None,
                                   **kwargs: dict) -> list:
        return await self.connector.get_my_recent_trades(symbol=symbol, since=since, limit=limit, **kwargs)

    async def cancel_order(self, order_id: str, symbol: str = None, **kwargs: dict) -> bool:
        return await self.connector.cancel_order(symbol=symbol, order_id=order_id, **kwargs)

    def get_trade_fee(self, symbol, order_type, quantity, price, taker_or_maker):
        return self.connector.get_trade_fee(symbol, order_type, quantity, price, taker_or_maker)

    def get_fees(self, symbol):
        return self.connector.get_fees(symbol)

    def get_pair_from_exchange(self, pair) -> str:
        return self.connector.get_pair_from_exchange(pair)

    def get_split_pair_from_exchange(self, pair) -> (str, str):
        return self.connector.get_split_pair_from_exchange(pair)

    def get_exchange_pair(self, pair) -> str:
        return self.connector.get_exchange_pair(pair)

    def get_pair_cryptocurrency(self, pair) -> str:
        return self.connector.get_pair_cryptocurrency(pair)

    def get_default_balance(self):
        return self.connector.get_default_balance()

    def get_rate_limit(self):
        return self.connector.get_rate_limit()

    async def switch_to_account(self, account_type: enums.AccountTypes):
        return await self.connector.switch_to_account(account_type=account_type)

    def parse_balance(self, balance):
        return personal_data.parse_decimal_portfolio(self.connector.parse_balance(balance))

    def parse_trade(self, trade):
        return self.connector.parse_trade(trade)

    def parse_order(self, order):
        return self.connector.parse_order(order)

    def parse_ticker(self, ticker):
        return self.connector.parse_ticker(ticker)

    def parse_ohlcv(self, ohlcv):
        return self.connector.parse_ohlcv(ohlcv)

    def parse_order_book(self, order_book):
        return self.connector.parse_order_book(order_book)

    def parse_order_book_ticker(self, order_book_ticker):
        return self.connector.parse_order_book_ticker(order_book_ticker)

    def parse_timestamp(self, data_dict, timestamp_key, default_value=None, ms=False):
        return self.connector.parse_timestamp(data_dict, timestamp_key, default_value=default_value, ms=ms)

    def parse_currency(self, currency):
        return self.connector.parse_currency(currency)

    def parse_order_id(self, order):
        return self.connector.parse_order_id(order)

    def parse_order_symbol(self, order):
        return self.connector.parse_order_symbol(order)

    def parse_status(self, status):
        return self.connector.parse_status(status)

    def parse_side(self, side):
        return self.connector.parse_side(side)

    def parse_account(self, account):
        return self.connector.parse_account(account)

    def clean_recent_trade(self, recent_trade):
        return self.connector.clean_recent_trade(recent_trade)

    def clean_trade(self, trade):
        return self.connector.clean_trade(trade)

    def clean_order(self, order):
        return self.connector.clean_order(order)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


""" Order class will represent an open order in the specified exchange
In simulation it will also define rules to be filled / canceled
It is also use to store creation & fill values of the order """
cimport octobot_trading.personal_data.portfolios.portfolio as portfolio
cimport octobot_trading.personal_data.portfolios.portfolio_manager as portfolio_manager

cdef class PortfolioValueHolder:
    cdef object logger
    cdef object config

    cdef public object portfolio_origin_value
    cdef public object portfolio_current_value

    cdef public dict last_prices_by_trading_pair
    cdef public dict origin_crypto_currencies_values
    cdef public dict current_crypto_currencies_values
    cdef public set initializing_symbol_prices

    cdef public portfolio.Portfolio origin_portfolio

    cdef set missing_currency_data_in_exchange

    cdef portfolio_manager.PortfolioManager portfolio_manager

    cpdef bint update_origin_crypto_currencies_values(self, str symbol, object mark_price)
    cpdef dict get_current_crypto_currencies_values(self)
    cpdef dict get_current_holdings_values(self)
    cpdef object get_origin_portfolio_current_value(self, bint refresh_values=*)
    cpdef object handle_profitability_recalculation(self, bint force_recompute_origin_portfolio)
    cpdef object convert_currency_value_using_last_prices(self, object quantity, str current_currency, str target_currency)
    # cpdef object get_currency_holding_ratio(self, str currency)

    cdef object _init_portfolio_values_if_necessary(self, bint force_recompute_origin_portfolio)
    cdef object _init_origin_portfolio_and_currencies_value(self)
    cdef object _update_portfolio_current_value(self, dict portfolio, dict currencies_values=*, bint fill_currencies_values=*)
    cdef void _fill_currencies_values(self, dict currencies_values)
    cdef dict _update_portfolio_and_currencies_current_value(self)
    cdef object _check_currency_initialization(self, str currency, object currency_value)
    cdef void _recompute_origin_portfolio_initial_value(self)
    cdef void _try_to_ask_ticker_missing_symbol_data(self, str currency, str symbol, str reversed_symbol)
    cdef void _ask_ticker_data_for_currency(self, list symbols_to_add)
    cdef void _inform_no_matching_symbol(self, str currency)
    cdef object _has_price_data(self, str symbol) # return object to propagate exceptions
    cdef _evaluate_config_crypto_currencies_and_portfolio_values(self,
                                                                dict portfolio,
                                                                bint ignore_missing_currency_data=*)
    cdef void _evaluate_config_currencies_values(self,
                                                 dict evaluated_pair_values,
                                                 set evaluated_currencies,
                                                 set missing_tickers)
    cdef void _evaluate_portfolio_currencies_values(self,
                                                    dict portfolio,
                                                    dict evaluated_pair_values,
                                                    set valuated_currencies,
                                                    set missing_tickers,
                                                    bint ignore_missing_currency_data)
    cdef object _evaluate_portfolio_value(self, dict portfolio, dict currencies_values=*)
    cdef bint _should_currency_be_considered(self, str currency, dict portfolio, bint ignore_missing_currency_data)
    # cdef object _evaluate_value(self, str currency, object quantity, bint raise_error=*)
    # cdef object _try_get_value_of_currency(self, str currency, object quantity, bint raise_error)
    # cdef object _get_currency_value(self, dict portfolio, str currency, dict currencies_values=*, bint raise_error=*)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest

from octobot_commons.tests.test_config import load_test_config

from octobot_trading.api.exchange import create_exchange_builder, \
    get_exchange_configurations_from_exchange_name, get_exchange_manager_from_exchange_name_and_id
from octobot_trading.exchanges.exchanges import Exchanges
from tests.exchanges import exchange_manager
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_create_exchange_builder():
    exchange_builder = create_exchange_builder(load_test_config(), "binance")
    assert exchange_builder


async def test_get_exchange_configurations_from_exchange_name(exchange_manager):
    Exchanges.instance().add_exchange(exchange_manager, None)
    assert get_exchange_configurations_from_exchange_name("binance")
    with pytest.raises(KeyError):
        get_exchange_configurations_from_exchange_name("bitmex")


async def test_get_exchange_manager_from_exchange_name_and_id(exchange_manager):
    Exchanges.instance().add_exchange(exchange_manager, None)
    assert get_exchange_manager_from_exchange_name_and_id(exchange_manager.exchange_name, exchange_manager.id)
    with pytest.raises(KeyError):
        get_exchange_manager_from_exchange_name_and_id(exchange_manager.exchange_name, "test")
    with pytest.raises(KeyError):
        get_exchange_manager_from_exchange_name_and_id("bitmex", exchange_manager.id)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_trading.constants as constants
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.personal_data.positions.channel.positions_updater as positions_updater


class PositionsUpdaterSimulator(positions_updater.PositionsUpdater):
    async def start(self):
        if not self._should_run():
            return
        await self.initialize()

    async def initialize(self) -> None:
        """
        Initialize positions and future contracts
        """
        await self.initialize_contracts()
        self.channel.exchange_manager.exchange_personal_data.positions_manager.positions_initialized = True
        self.logger = logging.get_logger(f"{self.__class__.__name__}[{self.channel.exchange_manager.exchange.name}]")
        await exchanges_channel.get_chan(constants.MARK_PRICE_CHANNEL, self.channel.exchange_manager.id) \
            .new_consumer(self.handle_mark_price)
        await exchanges_channel.get_chan(constants.FUNDING_CHANNEL, self.channel.exchange_manager.id) \
            .new_consumer(self.handle_funding_rate)
        self.channel.exchange_manager.exchange_personal_data.positions_manager.positions_initialized = True

    async def initialize_contracts(self) -> None:
        """
        Initialize exchange FutureContracts required to manage positions
        """
        for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
            self.channel.exchange_manager.exchange.create_pair_contract(
                pair=pair,
                current_leverage=constants.DEFAULT_SYMBOL_LEVERAGE,
                margin_type=constants.DEFAULT_SYMBOL_MARGIN_TYPE,
                contract_type=
                self.channel.exchange_manager.exchange_config.backtesting_exchange_config.future_contract_type
                if self.channel.exchange_manager.is_backtesting else constants.DEFAULT_SYMBOL_CONTRACT_TYPE,
                position_mode=constants.DEFAULT_SYMBOL_POSITION_MODE,
                maintenance_margin_rate=constants.DEFAULT_SYMBOL_MAINTENANCE_MARGIN_RATE,
                maximum_leverage=constants.DEFAULT_SYMBOL_MAX_LEVERAGE)

    async def handle_funding_rate(self, exchange: str,
                                  exchange_id: str,
                                  cryptocurrency: str,
                                  symbol: str,
                                  funding_rate,
                                  predicted_funding_rate,
                                  next_funding_time,
                                  timestamp):
        """
        Funding channel consumer callback
        """
        try:
            for symbol_position in self.channel.exchange_manager.exchange_personal_data.positions_manager. \
                    get_symbol_positions(symbol=symbol):
                if not symbol_position.is_idle() and symbol_position.symbol_contract.is_perpetual_contract():
                    async with self.channel.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.lock:
                        await self.channel.exchange_manager.exchange_personal_data.handle_portfolio_update_from_funding(
                            position=symbol_position, funding_rate=funding_rate
                        )
        except Exception as e:
            self.logger.exception(e, True, f"Fail to handle funding rate : {e}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchange_data.ohlcv.channel import ohlcv
from octobot_trading.exchange_data.ohlcv.channel.ohlcv import (
    OHLCVProducer,
    OHLCVChannel,
)
from octobot_trading.exchange_data.ohlcv.channel import ohlcv_updater
from octobot_trading.exchange_data.ohlcv.channel.ohlcv_updater import (
    OHLCVUpdater,
)

from octobot_trading.exchange_data.ohlcv.channel import ohlcv_updater_simulator
from octobot_trading.exchange_data.ohlcv.channel.ohlcv_updater_simulator import (
    OHLCVUpdaterSimulator,
)


__all__ = [
    "OHLCVUpdaterSimulator",
    "OHLCVProducer",
    "OHLCVChannel",
    "OHLCVUpdater",
]
